#!/usr/bin/env python
"""
#Import data from the online GMTKN24 database
(http://toc.uni-muenster.de/GMTKN/GMTKNmain.html).
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__changelog__ = '''
'''
__docformat__ = 'reStructuredText'


import itertools
import logging
import os
import os.path
import re
import sys
import tempfile
import time

from optparse import OptionParser

# for the `Gmtkn24` class
from mechanize import Browser, HTTPError
from BeautifulSoup import BeautifulSoup
from zipfile import ZipFile


PROG = os.path.splitext(os.path.basename(sys.argv[0]))[0]

## parse HTML tables

class HtmlTable(object):
    def __init__(self, html):
        """
        Initialize the `HtmlTable` object from HTML source.
        The `html` argument can be either the text source of
        an HTML fragment, or a BeautifulSource `Tag` object.
        """
        self._headers, self._rows = self.extract_html_table(html)

    def rows_as_dict(self):
        """
        Iterate over table rows, representing each row as a dictionary
        mapping header names into corresponding cell values.
        """
        for row in self._rows:
            yield dict([ (th[0], row[n])
                         for n,th in enumerate(self._headers) ])

    @staticmethod
    def extract_html_table(html):
        """
        Return the table in `html` text as a list of rows, each row being
        a tuple of the values found in table cells.
        """
        if not isinstance(html, BeautifulSoup):
            html = BeautifulSoup(html)
        table = html.find('table')
        # extract headers
        headers = [ ]
        for th in table.findAll('th'):
            text = str.join('', th.find(text=True)).strip()
            if th.has_key('colspan'):
                span = int(th['colspan'])
            else:
                span = 1
            headers.append((text, span))
        # extract rows and group cells according to the TH spans
        spans = [ s for (t,s) in headers ]
        rows = [ list(grouped(row, spans)) 
                      for row in HtmlTable.extract_html_table_rows(table) ]
        # all done
        return (headers, rows)

    @staticmethod
    def extract_html_table_rows(table):
        """
        Return list of rows (each row being a `tuple`) extracted 
        from HTML `Tag` element `table`.
        """
        return [ tuple([ str.join('', td.findAll(text=True)).strip()
                         for td in tr.findAll('td') ])
                 for tr in table.findAll('tr') ]

def grouped(iterable, pattern, container=tuple):
    """
    Iterate over elements in `iterable`, grouping them into 
    batches: the `n`-th batch has length given by the `n`-th
    item of `pattern`.  Each batch is cast into an object 
    of type `container`.

    Examples::

      >>> l = [0,1,2,3,4,5]
      >>> list(grouped(l, [1,2,3]))
      [0, (1, 2), (3, 4, 5)]
    """
    iterable = iter(iterable) # need a real iterator for tuples and lists
    for l in pattern:
        yield container(itertools.islice(iterable, l))


## interact with the online DB

class Gmtkn24(object):
    """
    Interact with the online web pages of GMTKN24.
    """
    BASE_URL = 'http://toc.uni-muenster.de/GMTKN/GMTKN24/'
    def __init__(self):
        # initialization
        self._browser = Browser()
        self._browser.set_handle_robots(False)
        self._subsets = self._list_subsets()

    _subset_link_re = re.compile("The (.+) subset")
    #_subset_link_re = re.compile("here")
    def _list_subsets(self):
        """Return dictionary mapping GMTKN24 subset names to download URLs."""
        html = BeautifulSoup(self._browser.open(Gmtkn24.BASE_URL + 'GMTKN24main.html'))
        links = html.findAll(name="a")
        result = { }
        for a in links:
	     if a.string is not None:
                match = Gmtkn24._subset_link_re.match(a.string)
                if match is not None:
                    print a
		    # if a subset has several names, add all of them
                    for name in match.group(1).split(' '):
                        if name == 'and':
                            continue
                        result[name] = Gmtkn24.BASE_URL + a['href']
        print result
	#result = ['google.com', 'cnn.com']
	return result

    def list(self):
        """Return dictionary mapping GMTKN24 subset names to download URLs."""
        return self._subsets

    def get_geometries(self, subset, output_dir='geometries'):
        """
        Download geometry files for the specified GMTKN24 subset,
        and save them into the 'geometries/' subdirectory of the
        current working directory.

        Return list of extracted molecules/filenames.
        """
        subset_url = self._subsets[subset]
        page = self._browser.open(subset_url)
        # must download the zip to a local file -- zipfiles are not stream-friendly ...
        geometries_url = self._browser.click_link(text_regex=re.compile("^Geometries"))
        (filename, headers) = self._browser.retrieve(geometries_url)
        logger.info("%s geometries downloaded into file '%s'", subset, filename)
        geometries_zip = ZipFile(filename, 'r')
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)
        molecules = self.get_molecule_names(subset)
        extracted = list()
        names = geometries_zip.namelist()
        for name in names:
            basename = os.path.basename(name)
            if basename not in molecules and basename != 'README':
                continue
            # zipfile's `extract` method preserves full pathname, 
            # so let's get the data from the archive and write
            # it in the file WE want...
            content = geometries_zip.read(name)
            output_path = os.path.join(output_dir, basename)
            output = open(output_path, 'w')
            output.write(content)
            output.close()
            if not ('README' == basename):
                extracted.append(basename)
            logger.info("Extracted '%s' into '%s'", basename, output_path)
        geometries_zip.close()
        return extracted

    def get_reference_data(self, subset):
        """
        Iterate over stoichiometry reference data in a given GMTKN24
        subset.  Each returned value is a pair `(r, d)`, where `r` is
        a dictionary mapping compound names (string) to their
        stoichiometric coefficient (integer), and `d` is a (float)
        number representing the total energy.
        """
        subset_url = self._subsets[subset]
        subset_page = self._browser.open(subset_url)
        if subset in ['BH76', 'BH76RC']:
            # special case
            self._browser.follow_link(text=("Go to the %s subset" % subset))
        refdata_page = self._browser.follow_link(text="Reference data")
        table = HtmlTable(refdata_page.read())
        for row in table.rows_as_dict():
            if subset == 'W4-08woMR':
                # The 16 entries marked with an asterisk (*) are not
                # part of the W4-08woMR subset.
                if row['#'] and row['#'][0].endswith('*'):
                    continue
            reactants = row['Systems']
            if len(reactants) == 0:
                continue # ignore null rows
            qtys = row['Stoichiometry']
            refdata = float(row['Ref.'][0])
            reaction = { }
            for n,sy in enumerate(reactants):
                if qtys[n] == '':
                    continue # skip null fields
                reaction[sy] = int(qtys[n])
            yield (reaction, refdata)

    def get_molecule_names(self, subset):
        """Return set of molecule names belonging in the specified subset."""
        # The only generic way to list molecule names seems to be:
        # take the systems names from the ref.data table.
        molecules = set()
        for reaction,data in self.get_reference_data(subset):
            for molecule in reaction:
                molecules.add(molecule)
        return molecules


## TURBOMOL to GAMESS conversion

class Struct(dict):
    """
    A `dict`-like object, whose keys can be accessed with the usual
    '[...]' lookup syntax, or with the '.' get attribute syntax.  
    Example::

      >>> x = Struct()
      >>> x.a = 1
      >>> x.has_key('a')
      True
      >>> x['a']
      1
      >>> x['b'] = 2
      >>> hasattr(x, 'b')
      True
      >>> x.b
      2

    An optional transformation function can be applied to the
    attribute/key names; for instance this is how the case-insensitive
    `UpcaseStruct` class (which see) is implemented:

      | >>> x = Struct(transform = lambda name: name.upper())
      | >>> x.a = 1
      | >>> x['A']
      | 1
    """
    def __init__(self, transform=None, **extra_args):
        if transform is not None:
            self.__dict__['_transform'] = transform
        else:
            def same(x):
                return x
            self.__dict__['_transform'] = same
        for k,v in extra_args.items():
             self[k] = v
    def __getattr__(self, key):
        return self[self.__dict__['_transform'](key)]
    def __getitem__(self, key):
        return dict.__getitem__(self, self.__dict__['_transform'](key))
    def __setattr__(self, key, val):
        self[self.__dict__['_transform'](key)] = val
    def __setitem__(self, key, val):
        dict.__setitem__(self, self.__dict__['_transform'](key), val)
    def has_key(self, key):
        return dict.has_key(self, self.__dict__['_transform'](key))


class UpcaseStruct(Struct):
    """
    A `dict`-like object, whose keys can be accessed with the usual
    '[...]' lookup syntax, or with the '.' get attribute syntax.  
    Example::

      >>> x = UpcaseStruct()
      >>> x.a = 1
      >>> x.has_key('a')
      True
      >>> x['a']
      1
      >>> x['b'] = 2
      >>> hasattr(x, 'b')
      True
      >>> x.b
      2

    Case of attribute/key names (with either syntax) is not
    significant: all attribute/key names will be canonicalized and
    stored as uppercase::

      >>> hasattr(x, 'B')
      True
      >>> x.has_key('B')
      True
      >>> x.B
      2
      >>> x.b is x.B
      True

    In particular, all attribute/key names will show as
    uppercase when printed::

      >>> x
      {'A': 1, 'B': 2}
    """
    __slots__ = [ ]
    def __init__(self, **extra_args):
        Struct.__init__(self, transform = lambda key: key.upper(), **extra_args)


class Atom(object):
    """
    Simple structure describing an atom in a molecule:
    a name, plus the three cartesian coordinates.
    """
    def __init__(self, name, x, y, z):
        self.name = Atom._DATA[name].name
        self.symbol = Atom._DATA[name].symbol
        self.nuclear_charge = Atom._DATA[name].nuclear_charge
        self.x = x
        self.y = y
        self.z = z

    # see: http://www.chem.qmul.ac.uk/iupac/AtWt/index.html
    _DATA = UpcaseStruct(
        H = Struct(name="Hydrogen", symbol="H", nuclear_charge=1.0, weight=1.00794),
        He = Struct(name="Helium", symbol="He", nuclear_charge=2.0, weight=4.002602),
        Li = Struct(name="Lithium", symbol="Li", nuclear_charge=3.0, weight=6.941),
        Be = Struct(name="Beryllium", symbol="Be", nuclear_charge=4.0, weight=9.012182),
        B = Struct(name="Boron", symbol="B", nuclear_charge=5.0, weight=10.811),
        C = Struct(name="Carbon", symbol="C", nuclear_charge=6.0, weight=12.0107),
        N = Struct(name="Nitrogen", symbol="N", nuclear_charge=7.0, weight=14.0067),
        O = Struct(name="Oxygen", symbol="O", nuclear_charge=8.0, weight=15.9994),
        F = Struct(name="Fluorine", symbol="F", nuclear_charge=9.0, weight=18.9984032),
        Ne = Struct(name="Neon", symbol="Ne", nuclear_charge=10.0, weight=20.1797),
        Na = Struct(name="Sodium", symbol="Na", nuclear_charge=11.0, weight=22.98976928),
        Mg = Struct(name="Magnesium", symbol="Mg", nuclear_charge=12.0, weight=24.3050),
        Al = Struct(name="Aluminium", symbol="Al", nuclear_charge=13.0, weight=26.9815386),
        Si = Struct(name="Silicon", symbol="Si", nuclear_charge=14.0, weight=28.0855),
        P = Struct(name="Phosphorus", symbol="P", nuclear_charge=15.0, weight=30.973762),
        S = Struct(name="Sulfur", symbol="S", nuclear_charge=16.0, weight=32.065),
        Cl = Struct(name="Chlorine", symbol="Cl", nuclear_charge=17.0, weight=35.453),
        Ar = Struct(name="Argon", symbol="Ar", nuclear_charge=18.0, weight=39.948),
        K = Struct(name="Potassium", symbol="K", nuclear_charge=19.0, weight=39.0983),
        Ca = Struct(name="Calcium", symbol="Ca", nuclear_charge=20.0, weight=40.078),
        Sc = Struct(name="Scandium", symbol="Sc", nuclear_charge=21.0, weight=44.955912),
        Ti = Struct(name="Titanium", symbol="Ti", nuclear_charge=22.0, weight=47.867),
        V = Struct(name="Vanadium", symbol="V", nuclear_charge=23.0, weight=50.9415),
        Cr = Struct(name="Chromium", symbol="Cr", nuclear_charge=24.0, weight=51.9961),
        Mn = Struct(name="Manganese", symbol="Mn", nuclear_charge=25.0, weight=54.938045),
        Fe = Struct(name="Iron", symbol="Fe", nuclear_charge=26.0, weight=55.845),
        Co = Struct(name="Cobalt", symbol="Co", nuclear_charge=27.0, weight=58.933195),
        Ni = Struct(name="Nickel", symbol="Ni", nuclear_charge=28.0, weight=58.6934),
        Cu = Struct(name="Copper", symbol="Cu", nuclear_charge=29.0, weight=63.546),
        Zn = Struct(name="Zinc", symbol="Zn", nuclear_charge=30.0, weight=65.38),
        Ga = Struct(name="Gallium", symbol="Ga", nuclear_charge=31.0, weight=69.723),
        Ge = Struct(name="Germanium", symbol="Ge", nuclear_charge=32.0, weight=72.64),
        As = Struct(name="Arsenic", symbol="As", nuclear_charge=33.0, weight=74.92160),
        Se = Struct(name="Selenium", symbol="Se", nuclear_charge=34.0, weight=78.96),
        Br = Struct(name="Bromine", symbol="Br", nuclear_charge=35.0, weight=79.904),
        Kr = Struct(name="Krypton", symbol="Kr", nuclear_charge=36.0, weight=83.798),
        Rb = Struct(name="Rubidium", symbol="Rb", nuclear_charge=37.0, weight=85.4678),
        Sr = Struct(name="Strontium", symbol="Sr", nuclear_charge=38.0, weight=87.62),
        Y = Struct(name="Yttrium", symbol="Y", nuclear_charge=39.0, weight=88.90585),
        Zr = Struct(name="Zirconium", symbol="Zr", nuclear_charge=40.0, weight=91.224),
        Nb = Struct(name="Niobium", symbol="Nb", nuclear_charge=41.0, weight=92.90638),
        Mo = Struct(name="Molybdenum", symbol="Mo", nuclear_charge=42.0, weight=95.96),
        Tc = Struct(name="Technetium", symbol="Tc", nuclear_charge=43.0, weight=98),
        Ru = Struct(name="Ruthenium", symbol="Ru", nuclear_charge=44.0, weight=101.07),
        Rh = Struct(name="Rhodium", symbol="Rh", nuclear_charge=45.0, weight=102.90550),
        Pd = Struct(name="Palladium", symbol="Pd", nuclear_charge=46.0, weight=106.42),
        Ag = Struct(name="Silver", symbol="Ag", nuclear_charge=47.0, weight=107.8682),
        Cd = Struct(name="Cadmium", symbol="Cd", nuclear_charge=48.0, weight=112.411),
        In = Struct(name="Indium", symbol="In", nuclear_charge=49.0, weight=114.818),
        Sn = Struct(name="Tin", symbol="Sn", nuclear_charge=50.0, weight=118.710),
        Sb = Struct(name="Antimony", symbol="Sb", nuclear_charge=51.0, weight=121.760),
        Te = Struct(name="Tellurium", symbol="Te", nuclear_charge=52.0, weight=127.60),
        I = Struct(name="Iodine", symbol="I", nuclear_charge=53.0, weight=126.90447),
        Xe = Struct(name="Xenon", symbol="Xe", nuclear_charge=54.0, weight=131.293),
        Cs = Struct(name="Caesium", symbol="Cs", nuclear_charge=55.0, weight=132.9054519),
        Ba = Struct(name="Barium", symbol="Ba", nuclear_charge=56.0, weight=137.327),
        La = Struct(name="Lanthanum", symbol="La", nuclear_charge=57.0, weight=138.90547),
        Ce = Struct(name="Cerium", symbol="Ce", nuclear_charge=58.0, weight=140.116),
        Pr = Struct(name="Praseodymium", symbol="Pr", nuclear_charge=59.0, weight=140.90765),
        Nd = Struct(name="Neodymium", symbol="Nd", nuclear_charge=60.0, weight=144.242),
        Pm = Struct(name="Promethium", symbol="Pm", nuclear_charge=61.0, weight=145),
        Sm = Struct(name="Samarium", symbol="Sm", nuclear_charge=62.0, weight=150.36),
        Eu = Struct(name="Europium", symbol="Eu", nuclear_charge=63.0, weight=151.964),
        Gd = Struct(name="Gadolinium", symbol="Gd", nuclear_charge=64.0, weight=157.25),
        Tb = Struct(name="Terbium", symbol="Tb", nuclear_charge=65.0, weight=158.92535),
        Dy = Struct(name="Dysprosium", symbol="Dy", nuclear_charge=66.0, weight=162.500),
        Ho = Struct(name="Holmium", symbol="Ho", nuclear_charge=67.0, weight=164.93032),
        Er = Struct(name="Erbium", symbol="Er", nuclear_charge=68.0, weight=167.259),
        Tm = Struct(name="Thulium", symbol="Tm", nuclear_charge=69.0, weight=168.93421),
        Yb = Struct(name="Ytterbium", symbol="Yb", nuclear_charge=70.0, weight=173.054),
        Lu = Struct(name="Lutetium", symbol="Lu", nuclear_charge=71.0, weight=174.9668),
        Hf = Struct(name="Hafnium", symbol="Hf", nuclear_charge=72.0, weight=178.49),
        Ta = Struct(name="Tantalum", symbol="Ta", nuclear_charge=73.0, weight=180.94788),
        W = Struct(name="Tungsten", symbol="W", nuclear_charge=74.0, weight=183.84),
        Re = Struct(name="Rhenium", symbol="Re", nuclear_charge=75.0, weight=186.207),
        Os = Struct(name="Osmium", symbol="Os", nuclear_charge=76.0, weight=190.23),
        Ir = Struct(name="Iridium", symbol="Ir", nuclear_charge=77.0, weight=192.217),
        Pt = Struct(name="Platinum", symbol="Pt", nuclear_charge=78.0, weight=195.084),
        Au = Struct(name="Gold", symbol="Au", nuclear_charge=79.0, weight=196.966569),
        Hg = Struct(name="Mercury", symbol="Hg", nuclear_charge=80.0, weight=200.59),
        Tl = Struct(name="Thallium", symbol="Tl", nuclear_charge=81.0, weight=204.3833),
        Pb = Struct(name="Lead", symbol="Pb", nuclear_charge=82.0, weight=207.2),
        Bi = Struct(name="Bismuth", symbol="Bi", nuclear_charge=83.0, weight=208.98040),
        Po = Struct(name="Polonium", symbol="Po", nuclear_charge=84.0, weight=209),
        At = Struct(name="Astatine", symbol="At", nuclear_charge=85.0, weight=210),
        Rn = Struct(name="Radon", symbol="Rn", nuclear_charge=86.0, weight=222),
        Fr = Struct(name="Francium", symbol="Fr", nuclear_charge=87.0, weight=223),
        Ra = Struct(name="Radium", symbol="Ra", nuclear_charge=88.0, weight=226),
        Ac = Struct(name="Actinium", symbol="Ac", nuclear_charge=89.0, weight=227),
        Th = Struct(name="Thorium", symbol="Th", nuclear_charge=90.0, weight=232.03806),
        Pa = Struct(name="Protactinium", symbol="Pa", nuclear_charge=91.0, weight=231.03588),
        U = Struct(name="Uranium", symbol="U", nuclear_charge=92.0, weight=238.02891),
        Np = Struct(name="Neptunium", symbol="Np", nuclear_charge=93.0, weight=237),
        Pu = Struct(name="Plutonium", symbol="Pu", nuclear_charge=94.0, weight=244),
        Am = Struct(name="Americium", symbol="Am", nuclear_charge=95.0, weight=243),
        Cm = Struct(name="Curium", symbol="Cm", nuclear_charge=96.0, weight=247),
        Bk = Struct(name="Berkelium", symbol="Bk", nuclear_charge=97.0, weight=247),
        Cf = Struct(name="Californium", symbol="Cf", nuclear_charge=98.0, weight=251),
        Es = Struct(name="Einsteinium", symbol="Es", nuclear_charge=99.0, weight=252),
        Fm = Struct(name="Fermium", symbol="Fm", nuclear_charge=100.0, weight=257),
        Md = Struct(name="Mendelevium", symbol="Md", nuclear_charge=101.0, weight=258),
        No = Struct(name="Nobelium", symbol="No", nuclear_charge=102.0, weight=259),
        Lr = Struct(name="Lawrencium", symbol="Lr", nuclear_charge=103.0, weight=262),
        Rf = Struct(name="Rutherfordium", symbol="Rf", nuclear_charge=104.0, weight=265),
        Db = Struct(name="Dubnium", symbol="Db", nuclear_charge=105.0, weight=268),
        Sg = Struct(name="Seaborgium", symbol="Sg", nuclear_charge=106.0, weight=271),
        Bh = Struct(name="Bohrium", symbol="Bh", nuclear_charge=107.0, weight=272),
        Hs = Struct(name="Hassium", symbol="Hs", nuclear_charge=108.0, weight=270),
        Mt = Struct(name="Meitnerium", symbol="Mt", nuclear_charge=109.0, weight=276),
        Ds = Struct(name="Darmstadtium", symbol="Ds", nuclear_charge=110.0, weight=281),
        Rg = Struct(name="Roentgenium", symbol="Rg", nuclear_charge=111.0, weight=280),
        Cn = Struct(name="Copernicium", symbol="Cn", nuclear_charge=112.0, weight=285),
        Uut = Struct(name="Ununtrium", symbol="Uut", nuclear_charge=113.0, weight=284),
        Uuq = Struct(name="Ununquadium", symbol="Uuq", nuclear_charge=114.0, weight=289),
        Uup = Struct(name="Ununpentium", symbol="Uup", nuclear_charge=115.0, weight=288),
        Uuh = Struct(name="Ununhexium", symbol="Uuh", nuclear_charge=116.0, weight=293),
        Uuo = Struct(name="Ununoctium", symbol="Uuo", nuclear_charge=118.0, weight=294),
        )

_whitespace_re = re.compile(r'\s+', re.X)

def read_turbomole_coords(input):
    """
    Import a TURBOMOLE format file.

    Argument `input` can be either a file-like object, providing a
    `read` method that is used to read its full contents, or a
    string, containing the path name of a file to read.
    """
    if hasattr(input, 'read'):
        lines = iter(input.read().split('\n'))
    else:
        f = open(input, 'rU')
        lines = iter(f.read().split('\n'))
        f.close()
    result = [ ]
    try:
        while True:
            line = lines.next().strip()
            # skip empty lines
            if line == '':
                continue
            if line.lower().startswith('$coord'):
                while True:
                    coord_line = lines.next().strip()
                    if coord_line.lower().startswith('$end'):
                        break
                    x, y, z, symbol = _whitespace_re.split(coord_line)
                    result.append(Atom(symbol, float(x), float(y), float(z)))
    except StopIteration:
        pass
    return result
                

def turbomol_to_gamess(input_file_name, output_file_name, template_file_name, **extra_args):
    """
    Read TURBOMOL coordinates and output a GAMESS file using the
    specified template.
    """
    atoms = read_turbomole_coords(input_file_name)
    # construct GAMESS $DATA section
    data = str.join('\n',
                    [ ('%-10s %3.1f  %+2.8f  %+2.8f  %+2.8f' 
                       % (a.symbol, a.nuclear_charge, a.x, a.y, a.z))
                      for a in atoms ])
    # get GAMESS '.inp' template
    f = open(template_file_name, 'rU')
    template = f.read()
    f.close()
    # write output
    f = open(output_file_name, 'w')
    extra_args['NAME'] = os.path.basename(input_file_name)
    extra_args['ATOMS'] = data
    f.write(template % extra_args)
    f.close()


## parse command-line options

cmdline = OptionParser(PROG + " [options] ACTION [SUBSET]",
                       description="""
Imports geometries and reference data from the GMTKN24 database into
GAMESS `.inp` format.  The actual behavior of this script depends on
the ACTION pargument.

If ACTION is "list", a list of all available subsets is printed on the
standard output.

If ACTION is "import", the specified subset is imported into the
current directory.  If no SUBSET is specified, *all* available subsets
are converted.
""")
cmdline.add_option("-d", "--destination-dir", dest="destdir", metavar="DIR",
                   default=os.path.join(os.getcwd(), 'wiki'),
                   help="Import data into the specified directory."
                   " Default: %default")
cmdline.add_option("-t", "--template", dest="template", metavar="TEMPLATE",
                   default="etc/gmtkn24.inp.template",
                   help="Use the specified template file for constructing a GAMESS '.inp'"
                   " input file from GMTKN24 molecule geometries.  (Default: '%default')"
                   )
cmdline.add_option("-v", "--verbose", dest="verbose", action="count", default=0, 
                   help="Verbosely report about program operation and progress."
                   )
(options, args) = cmdline.parse_args()


## main

logging.basicConfig(level=max(1, logging.ERROR - 10 * options.verbose),
                    format='%(name)s: %(message)s')
logger = logging.getLogger(PROG)


def get_required_argument(args, argno=1, argname='SUBSET'):
    try:
        return args[argno]
    except:
        logger.critical("Missing required argument %s."
                        " Aborting now; type '%s --help' to get usage help.", 
                        argname, PROG)
        raise
def get_optional_argument(args, argno=1, argname="SUBSET", default=None):
    try:
        return args[argno]
    except:
        logger.info("Missing argument %s, using default '%s'", argname, default)
        return default


if "__main__" == __name__:
    if len(args) < 1:
        logger.critical("%s: Need at least the ACTION argument."
                        " Type '%s --help' to get usage help."
                        % (PROG, PROG))
        sys.exit(1)

    try:
        if 'import' == args[0]:
            gmtkn24 = Gmtkn24()
            subset = get_optional_argument(args, 1, "SUBSET", None)
            subset_urls = gmtkn24.list()
            if subset:
                subsets = [ subset ]
            else:
                subsets = [ name for name in subset_urls.keys() ]
            tempdir = tempfile.mkdtemp(prefix=PROG)
            destdir = options.destdir
            if not os.path.isabs(destdir):
                destdir = os.path.join(os.getcwd(), destdir)
            template = options.template
            for subset in subsets:
                url = subset_urls[subset]
                if not os.path.exists(os.path.join(destdir, subset)):
                    os.mkdir(os.path.join(destdir, subset))
                ## download geometries and convert to GAMESS .inp format
                logger.info("Downloading %s geometries into '%s' ...", subset, destdir)
                molecules = gmtkn24.get_geometries(subset, tempdir)
                # load GAMESS parameters SCFTYP, ICHARG, MULT with default values
                params = { }
                for molecule in molecules:
                    params[molecule] = { 'SCFTYP':'RHF', 'ICHARG':0, 'MULT':1 }
                readme = os.path.join(tempdir, 'README')
                if os.path.exists(readme):
                    charge_or_unpaired_re = re.compile(r'(?P<molecule>[a-z0-9+_,-]+) \s+'
                                                       r'\((?P<charge_or_unpaired>[+-]?[0-9]+)\)', 
                                                       re.I|re.X)
                    readme_file = open(readme, 'r')
                    for line in readme_file:
                        match = charge_or_unpaired_re.match(line)
                        if match:
                            molecule = match.group('molecule')
                            if molecule not in molecules:
                                logger.warning("Ignoring molecule '%s': mentioned by README file,"
                                               " but not listed in %s stoichiometry data",
                                               molecule, subset)
                                continue
                            charge_or_unpaired = match.group('charge_or_unpaired')
                            if charge_or_unpaired.startswith('+') or charge_or_unpaired.startswith('-'):
                                # signed value, must be charge
                                charge = int(charge_or_unpaired)
                                params[molecule]['ICHARG'] = charge
                                params[molecule]['SCFTYP'] = 'UHF'
                                logger.info("Will use ICHARG=%s for molecule %s", charge, molecule)
                            else:
                                # unsigned value, must be unpaired electron no.
                                mult = 1 + int(charge_or_unpaired)
                                params[molecule]['MULT'] = mult
                                if mult != 1:
                                    params[molecule]['SCFTYP'] = 'UHF'
                                logger.info("Will use MULT=%s for molecule %s", mult, molecule)
                    readme_file.close()
                else:
                    logger.info("No README file accompanies geometries; will use default: SCFTYP=RHF, ICHARG=0 and MULT=1")
                for molecule in molecules:
                    destfile = os.path.join(destdir, subset, molecule + '.inp')
                    logger.info("Generating GAMESS input file '%s' from geometry '%s' and template '%s'...",
                                destfile, molecule, template)
                    turbomol_to_gamess(tempdir + '/' + molecule, destfile, template, 
                                       **params[molecule])
                ## write out reference data in `.csv` format
                refdata = list(gmtkn24.get_reference_data(subset))
                refdata_csv = os.path.join(destdir, subset, 'refdata.csv')
                output = open(refdata_csv, 'w')
                maxwidth = max([ len(reaction) for reaction,energy in refdata ])
                output.write("#;" + 
                             "Systems" + (';' * maxwidth) +
                             "Stoichiometry" + (';' * maxwidth) +
                             "Ref.\n")
                for seqno, data in enumerate(refdata):
                    reaction, energy = data
                    width = len(reaction)
                    systems = str.join(";", reaction.keys() + ([" "] * (maxwidth - width)))
                    stoichiometry = str.join(";", [ str(coeff) for coeff in reaction.values() ]
                                             + ([" "] * (maxwidth - width)))
                    output.write("%d;%s;%s;%.2f\n" % (seqno, systems, stoichiometry, energy))
                output.close()
                ## finally, write subset description page
                output = open(os.path.join(destdir, subset, 'index.mdwn'), 'w')
                output.write("""
[[!tag subset %(database)s %(subset)s]]

Reference data
--------------

For each reaction, the relevant systems' names, the stoichiometry and the reference value are given.
The systems' names refer to the geometry files (see section "GAMESS input" below).
Negative values in the stoichiometry columns refer to reactants, positive values to products. 

<!-- DO NO CHANGE THE "table" LINE BELOW: it generates the table of ref.data
     If you want to edit/replace the reference data, 
     use the `Attachments` button and replace the `refdata.csv` file -->
[[!table class="refdata" delimiter=";" file="subsets/%(subset)s/refdata.csv"]]


Reference GAMESS input
----------------------
<!-- DO NO CHANGE THE "map" LINE BELOW: it generates the list of GAMESS input files.
     If you want to edit/replace the `.inp` files, use the `Attachments` button. -->
[[!map pages="glob(subsets/%(subset)s/*.inp)"]]


--------------------------------
These data have initially been imported from the online
[%(database)s database](%(database_url)s)
""" % { 'database':'GMTKN24',
        'database_url':url,
        'subset':subset, 
        # XXX: URLs need to be relative to the Wiki top-level?
        'refdata_csv':os.path.join(subset, os.path.basename(refdata_csv))
        })
                output.close()
                                 

        elif 'list' == args[0]:
            subset = get_optional_argument(args, 1, "SUBSET")
            if subset:
                print ("Molecules in the %s subset:" % subset)
                for name in Gmtkn24().get_molecule_names(subset):
                    print ("  %s" % name)
            else:
                print ("Available subsets of GMTKN24:")
                ls = Gmtkn24().list()
                for name, url in ls.items():
                    print ("  %s --> %s" % (name, url))
            
        elif 'refdata' == args[0]:
            subset = get_required_argument(args, 1, "SUBSET")
            for r,d in Gmtkn24().get_reference_data(subset):
                print ("%s = %.3f" 
                       % (str.join(' + ', 
                                   [ ("%d*%s" % (qty, sy)) for sy,qty in r.items() ]), 
                          d))

        elif 'doctest' == args[0]:
            import doctest
            doctest.testmod(name="gmtkn24",
                            optionflags=doctest.NORMALIZE_WHITESPACE)

        else:
            logger.critical("Unknown ACTION word '%s'."
                            " Type '%s --help' to get usage help."
                            % (args[0], PROG))
        sys.exit(1)

    except HTTPError, x:
        logger.critical("HTTP error %d requesting page: %s" % (x.code, x.msg))
        sys.exit(1)




#! /usr/bin/env python
#
"""
Unit tests for the `gc3libs.optimizer.drivers` module.
"""
# Copyright (C) 2011, 2012, 2013, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'

import os
import sys
import logging
import tempfile
import shutil

from nose.tools import raises, assert_true
from nose.plugins.skip import SkipTest

import cli.test

import numpy as np

import gc3libs
from gc3libs import Application, configure_logger
from gc3libs.cmdline import SessionBasedScript

# optimizer specific imports
from gc3libs.optimizer import draw_population
from gc3libs.utils import update_parameter_in_file
from gc3libs.optimizer.drivers import ParallelDriver, SequentialDriver
from gc3libs.optimizer.dif_evolution import DifferentialEvolutionAlgorithm
from gc3libs.optimizer.extra import print_stats, log_stats, plot_population

class TestSequentialDriver(object):

    def setUp(self):
        # Create a temporary stage directory
        self.temp_stage_dir = tempfile.mkdtemp(prefix = 'SequentialDriver_Rosenbrock_')        

    def tearDown(self):
        # Remove Rosenbrock output
        shutil.rmtree(self.temp_stage_dir)

    def test_SequentialDriver_with_rosenbrock(self):
        """Test :class:`gc3libs.optimizer.drivers.SequentialDriver
        """
        
        # Test parameters
        magic_seed = 100
        dim = 2
        pop_size = 100
        lower_bounds = -2 * np.ones(dim)
        upper_bounds = +2 * np.ones(dim)
        prob_cross = 0.8
             

        configure_logger(level=logging.CRITICAL)
        log = logging.getLogger("gc3.gc3libs")        
        
        initial_pop = draw_population(lower_bds = lower_bounds, upper_bds = upper_bounds, dim = dim, size = pop_size,
                                      in_domain = self.rosenbrock_sample_filter, seed = magic_seed)

        algo = DifferentialEvolutionAlgorithm(
            initial_pop = initial_pop,
            de_step_size = 0.85,# DE-stepsize ex [0, 2]
            prob_crossover = prob_cross, # crossover probabililty constant ex [0, 1]
            itermax = 1000,      # maximum number of iterations (generations)
            dx_conv_crit = None, # stop when variation among x's is < this
            y_conv_crit = 1e-5, # stop when ofunc < y_conv_crit
            de_strategy = 'DE_local_to_best',
            logger = log,
            in_domain=self.rosenbrock_sample_filter,
            seed = magic_seed,
            after_update_opt_state=[print_stats, log_stats]
                                    #, plot_population(temp_stage_dir)]
            )
        assert algo.de_step_size == 0.85
        assert algo.prob_crossover == prob_cross
        assert algo.itermax == 1000
        assert algo.dx_conv_crit == None
        assert algo.y_conv_crit == 1e-5
        assert algo.de_strategy == 'DE_local_to_best'
        assert algo.logger == log

        opt = SequentialDriver(algo, target_fn=self.rosenbrock_fn, fmt="%12.8f")
        assert opt.target_fn == self.rosenbrock_fn

        # run the Diff.Evo. algorithm
        opt.de_opt()

        assert algo.has_converged()
        assert (algo.best_y - 0.) < algo.y_conv_crit
        assert (algo.best_x[0] - 1.) < 1e-3
        assert (algo.best_x[1] - 1.) < 1e-3
        
    @staticmethod
    def rosenbrock_fn(vectors):
        result = []
        for vector in vectors:
            #---Rosenbrock saddle-------------------------------------------
            F_cost = 100 * ( vector[1] - vector[0]**2 )**2 + ( 1 - vector[0] )**2
    
            result.append(F_cost)
        return np.array(result)
    
    @staticmethod
    def rosenbrock_sample_filter(pop):
        '''
        Sample filter function.
        In optimum x[0] + x[1] = 2.
        '''
        filter_pop_sum = 3.                # x[0] + x[1] <= filter_pop_sum   
        return [ x[0] + x[1] <= filter_pop_sum for x in pop ]
    

class TestParallelDriver(cli.test.FunctionalTest):
    CONF="""
[resource/localhost_test]
type=shellcmd
transport=local
time_cmd=/usr/bin/time
max_cores=2
max_cores_per_job=2
max_memory_per_core=2
max_walltime=2
architecture=x64_64
auth=noauth
enabled=True
resourcedir=%s

[auth/noauth]
type=none
"""
    def __init__(self, *args, **extra_args):
        cli.test.FunctionalTest.__init__(self, *args, **extra_args)
        self.scriptdir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '../../../examples/optimizer/rosenbrock')

    def setUp(self):
        cli.test.FunctionalTest.setUp(self)
        
        # Create stage_dir to run optimization in. 
        self.temp_stage_dir = tempfile.mkdtemp(prefix = 'ParallelDriver_Rosenbrock_')
        
        # Create a sample user config file. 
        (fd, cfgfile) = tempfile.mkstemp()
        resourcedir = cfgfile + '.d'
        f = os.fdopen(fd, 'w+')
        f.write(TestParallelDriver.CONF % resourcedir)
        f.close()
        sys.argv += ['--config-files', cfgfile, '-r', 'localhost_test']
        self.files_to_remove = [cfgfile, resourcedir]

        self.cfg = gc3libs.config.Configuration()
        self.cfg.merge_file(cfgfile)

        self.core = gc3libs.core.Core(self.cfg)
        self.backend = self.core.get_backend('localhost_test')

        # Create base dir with sample parameter file
        self.temp_base_dir = os.path.join(self.temp_stage_dir, 'base')
        os.mkdir(self.temp_base_dir)
        para_file = open(os.path.join(self.temp_base_dir, 'parameters.in'), 'w')
        para_file.write('x1    100.0\nx2    50.0\n')


    def cleanup_file(self, fname):
        self.files_to_remove.append(fname)

    def tearDown(self):
        cli.test.FunctionalTest.tearDown(self)
        
        for fname in self.files_to_remove:
            if os.path.isdir(fname):
                shutil.rmtree(fname)
            elif os.path.exists(fname):
                os.remove(fname)

        # Remove Rosenbrock output
        shutil.rmtree(self.temp_stage_dir)

    def test_ParallelDriver(self):
        """Test :class:`gc3libs.optimizer.drivers.ParallelDriver`

        The script is found in ``examples/optimizer/rosenbrock/opt_rosenbrock.py``
        """
        # Run Rosenbrock
        result = self.run_script('python',
                                 os.path.join(self.scriptdir, 'opt_rosenbrock.py'),
                                 '-C', '1',
                                 '-r', 'localhost',
                                 '--pop_size', 5,
                                 '--y_conv_crit', 1000.0,
                                 '--path_to_stage_dir', self.temp_stage_dir, 
                                 '--path_to_base_dir', self.temp_base_dir,
                                 '--path_to_executable', os.path.join(os.path.dirname(__file__), 'fitness_func', 'rosenbrock.py')
                               #  ,'-vvvv'
                                 )
        
        assert_true(result.stderr.find('Converged:'))


## main: run tests

if "__main__" == __name__:
    import nose
    nose.runmodule()



#! /usr/bin/env python
#
#   gcelljunction.py -- GC3Pie front-end for running the
#   "tricellular_junction" code by T. Aegerter
#
#   Copyright (C) 2014 GC3, University of Zurich
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for runniung multiple `tricellular_junction` instances.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gcelljunction --help`` for program usage instructions.
"""
__version__ = '2.2.1 version (SVN $Revision$)'
# summary of user-visible changes
__changelog__ = """
  2014-03-03:
    * Initial release, forked off the ``gmhc_coev`` sources.
"""
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == '__main__':
    import gcelljunction
    gcelljunction.GCellJunctionScript().run()


# std module imports
import csv
import glob
import os
import re
import sys
import time
from pkg_resources import Requirement, resource_filename

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript
from gc3libs.compat._collections import defaultdict
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask


## custom application class

class GCellJunctionApplication(Application):
    """
    Custom class to wrap the execution of the ``tricellular_junction``
    program by T. Aegerter.
    """

    application_name = 'tricellular_junction'

    def __init__(self, sim_no, executable=None, **extra_args):
        self.sim_no = sim_no
        wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                       "gc3libs/etc/gcelljunction_wrapper.sh")
        inputs = { wrapper_sh:os.path.basename(wrapper_sh) }
        extra_args.setdefault('requested_cores',        1)
        extra_args.setdefault('requested_memory',       3*GB)
        extra_args.setdefault('requested_architecture', Run.Arch.X86_64)
        extra_args.setdefault('requested_walltime',     60*Duration.days)
        # command-line parameters to pass to the tricellular_junction_* program
        self.sim_no = sim_no
        if executable is not None:
            # use the specified executable
            exename = os.path.basename(executable)
            executable_name = './' + exename
            inputs[executable] = exename
            exe_opts = ['-x', executable_name]
        else:
            # assume one is installed in the VM
            executable_name = 'tricellular_junctions'
            exe_opts = [ ]
        Application.__init__(
            self,
            arguments=['./' + os.path.basename(wrapper_sh)] + exe_opts + [ '--', sim_no ],
            inputs = inputs,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'tricellular_junctions.log',
            join=True,
            **extra_args)


class GCellJunctionTask(RetryableTask, gc3libs.utils.Struct):
    """
    Retry execution of a `GCellJunctionApplication` if it fails.
    """
    def __init__(self, sim_no, executable=None, **extra_args):
        self.sim_no = sim_no
        RetryableTask.__init__(
            self,
            # actual computational job
            GCellJunctionApplication(sim_no, executable, **extra_args),
            # keyword arguments
            **extra_args)

    _CHECK_LINES = 5
    def update_state(self, **extra_args):
        super(GCellJunctionTask, self).update_state(**extra_args)
        if self.execution.state == Run.State.RUNNING:
            try:
                estimated_size = gc3libs.Default.PEEK_FILE_SIZE * self._CHECK_LINES
                with self.task.peek('stdout', offset=-estimated_size, size=estimated_size) as fd:
                    # drop first and last lines, as they may be partial
                    lines = fd.readlines()[1:-1]
                    # gc3libs.log.debug(
                    #     "Lines read from remote output: %s",
                    #     str.join(' ', [("<<%s>>" % ln) for ln in lines]))
                    for line in reversed(lines):
                        line = line.strip()
                        if line != '':
                            self.execution.info = line
                            break
            except Exception, err:
                gc3libs.log.warning(
                    "Ignored error while updating state of Task %s: %s: %s",
                    self, err.__class__.__name__, err)


## main script class

class GCellJunctionScript(SessionBasedScript):
    """
Read the specified INPUT ``.csv`` files and submit jobs according
to the content of those files.  Job progress is monitored and, when a
job is done, its ``data/`` and ``data4/`` output directories are
retrieved back into the same directory where the executable file is
(this can be overridden with the ``-o`` option).

The ``gcelljunction`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``gcelljunction`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            input_filename_pattern = '*.csv',
            application = GCellJunctionApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GCellJunctionTask,
            )


    def setup_options(self):
        self.add_param("-x", "--executable", metavar="PATH",
                       dest="executable", default=None,
                       help="Path to the `tricellular_junctions` executable file.")
        # change default for the memory/walltime options
        self.actions['memory_per_core'].default = 4*Memory.GB
        self.actions['wctime'].default = '60 days'


    def make_directory_path(self, pathspec, jobname):
        # XXX: Work around SessionBasedScript.process_args() that
        # apppends the string ``NAME`` to the directory path.
        # This is really ugly, but the whole `output_dir` thing needs to
        # be re-thought from the beginning...
        if pathspec.endswith('/NAME'):
            return pathspec[:-len('/NAME')]
        else:
            return pathspec


    def before_main_loop(self):
        # XXX: should this be done with `make_controller` instead?
        self._controller.retrieve_running = True
        self._controller.retrieve_overwrites = True
        self._controller.retrieve_changed_only = True


    def new_tasks(self, extra):
        # how many iterations are we already computing (per parameter set)?
        iters = defaultdict(int)
        for task in self.session:
            name, instance = task.jobname.split('#')
            iters[name] = max(iters[name], int(instance))

        for path in self.params.args:
            if path.endswith('.csv'):
                try:
                    inputfile = open(path, 'r')
                except (OSError, IOError), ex:
                    self.log.warning("Cannot open input file '%s': %s: %s",
                                     path, ex.__class__.__name__, str(ex))
                for lineno, line in enumerate(inputfile):
                    line = line.strip()
                    # ignore blank and comment lines (those that start with '#')
                    if len(line) == 0 or line.startswith('#'):
                        continue
                    try:
                        sim_no = int(line)
                    except ValueError:
                        self.log.error("Wrong format in line %d of file '%s':"
                                       " need 1 integer value (`SimNo`),"
                                       " but actually got '%s'."
                                       " Ignoring input line, fix it and re-run.",
                                       lineno+1, path, line)
                        continue # with next `row`
                    # extract parameter values
                    basename = ('tricellular_junction_%d' % (sim_no,))

                    # prepare job(s) to submit
                    already = len([ task for task in self.session if task.sim_no == sim_no ])
                    kwargs = extra.copy()
                    base_output_dir = kwargs.pop('output_dir', self.params.output)
                    jobname=('%s#%d' % (basename, already+1))
                    yield GCellJunctionTask(
                        sim_no,
                        executable=self.params.executable,
                        jobname=jobname,
                        output_dir=os.path.join(base_output_dir, jobname),
                        **kwargs)

            else:
                self.log.error("Ignoring input file '%s': not a CSV file.", path)



#! /usr/bin/env python
#
"""
Job control on SLURM clusters (possibly connecting to the front-end via SSH).
"""
# Copyright (C) 2012-2013 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.2.1 version (SVN $Revision$)'


import datetime
import os
import posixpath
import random
import re
import sys
import tempfile
import time

from gc3libs.compat._collections import defaultdict

from gc3libs import log, Run
from gc3libs.backends import LRMS
import gc3libs.backends.batch as batch
import gc3libs.exceptions
from gc3libs.quantity import Memory, bytes, kB, MB, GB
from gc3libs.quantity import Duration, seconds, minutes, hours
import gc3libs.backends.transport as transport
import gc3libs.utils as utils  # first, to_bytes
from gc3libs.utils import same_docstring_as


# environmental variables:
#        SLURM_TIME_FORMAT Specify the format used to report time
#        stamps. A value of standard, the default value, generates
#        output in the form "year-month-dateThour:minute:second".

# stat cmd: squeue --noheader --format='%i^%T^%r^%R'  -j jobid1,jobid2,...
#   %i: job id
#
#   %T: Job state, extended form: PENDING, RUNNING, SUSPENDED,
#       CANCELLED, COMPLETING, COMPLETED, CONFIGURING, FAILED, TIMEOUT,
#       PREEMPTED, and NODE_FAIL.
#
#   %R: For pending jobs: the reason a job is waiting for execution is
#       printed within parenthesis. For terminated jobs with failure: an
#       explanation as to why the job failed is printed within
#       parenthesis.  For all other job states: the list of allocate
#       nodes.
#   %r: reason a job is in its current state


## data for parsing SLURM commands output

# regexps for extracting relevant strings

# `sbatch` examples:
#
# $ sbatch -N4 myscript
# salloc: Granted job allocation 65537
#
# $ sbatch -N4 <<EOF
# > #!/bin/sh
# > srun hostname |sort
# > EOF
# sbatch: Submitted batch job 65541
#
_sbatch_jobid_re = re.compile(
    r'(sbatch:\s*)?(Granted job allocation|Submitted batch job)'
    ' (?P<jobid>\d+)')

# `squeue` examples:
#
#    $ squeue --noheader --format='%i|%T|%r|%R' -j 2,3
#    2|PENDING|Resources|(Resources)
#    3|PENDING|Resources|(Resources)
#


## code

def count_jobs(squeue_output, whoami):
    """
    Parse SLURM's ``squeue`` output and return a quadruple `(R, Q, r,
    q)` where:

      * `R` is the total number of running jobs (from any user);
      * `Q` is the total number of queued jobs (from any user);
      * `r` is the number of running jobs submitted by user `whoami`;
      * `q` is the number of queued jobs submitted by user `whoami`

    The `squeue_output` must contain the results of an invocation of
    ``squeue --noheader --format='%i^%T^%u^%U^%r^%R'``.
    """
    total_running = 0
    total_queued = 0
    own_running = 0
    own_queued = 0
    for line in squeue_output.split('\n'):
        if line == '':
            continue
        # the choice of format string makes it easy to parse squeue output
        jobid, state, username, uid, reason, nodelist = line.split('^')
        if state in ['RUNNING', 'COMPLETING']:
            total_running += 1
            if username == whoami:
                own_running += 1
        # XXX: State CONFIGURING is described in the squeue(1) man
        # page as "Job has been allocated resources, but are waiting
        # for them to become ready for use (e.g. booting).".  Should
        # it be classified as "running" instead?
        elif state in ['PENDING', 'CONFIGURING']:
            total_queued += 1
            if username == whoami:
                own_queued += 1
    return (total_running, total_queued, own_running, own_queued)


class SlurmLrms(batch.BatchSystem):
    """
    Job control on SLURM clusters (possibly by connecting via SSH to a
    submit node).
    """

    _batchsys_name = 'SLURM'

    def __init__(self, name,
                 # this are inherited from the base LRMS class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 auth,  # ignored if `transport` is 'local'
                 # these are inherited from `BatchSystem`
                 frontend, transport,
                 # these are specific to this backend
                 # (Note that optional arguments to the `BatchSystem` class, e.g.:
                 #     keyfile=None, accounting_delay=15,
                 # are collected into `extra_args` and should not be explicitly
                 # spelled out in this signature.)
                 **extra_args):

        # init base class
        batch.BatchSystem.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth,
            frontend, transport,
            **extra_args)

        # backend-specific setup
        self.sbatch = self._get_command_argv('sbatch')

        # SLURM commands
        self._scancel = self._get_command('scancel')
        self._squeue = self._get_command('squeue')
        self._sacct = self._get_command('sacct')

    def _parse_submit_output(self, output):
        return self.get_jobid_from_submit_output(output, _sbatch_jobid_re)

    # submit cmd: sbatch --job-name="jobname" --mem-per-cpu="MBs" --input="filename" --output="filename" --no-requeue -n "number of slots" --cpus-per-task=1 --time="minutes" script.sh
    #
    # You can only submit scripts with `sbatch`, attempts to directly run a command fail with this error message:
    #
    #    $ sbatch sleep 180
    #    sbatch: error: This does not look like a batch script.  The first
    #    sbatch: error: line must start with #! followed by the path to an interpreter.
    #    sbatch: error: For instance: #!/bin/sh
    #
    # If the server is not running or unreachable, the following error is displayed:
    #
    #    $ sbatch -n 1 --no-requeue /tmp/sleep.sh
    #    sbatch: error: Batch job submission failed: Unable to contact slurm controller (connect failure)
    #
    # Note: SLURM has a flexible way of assigning CPUs/cores/etc. to a job; here we treat all execution slots as equal and map 1 task to 1 core/thread.
    #
    # Acceptable time formats include "minutes",  "minutes:seconds", "hours:minutes:seconds", "days-hours", "days-hours:minutes" and "days-hours:minutes:seconds".
    #
    def _submit_command(self, app):
        sbatch_argv, app_argv = app.sbatch(self)
        return (str.join(' ', sbatch_argv), str.join(' ', app_argv))

    # stat cmd: squeue --noheader --format='%i^%T^%u^%U^%r^%R'  -j jobid1,jobid2,...
    #   %i: job id
    #   %T: Job  state,  extended  form: PENDING, RUNNING, SUSPENDED, CANCELLED, COMPLETING, COMPLETED, CONFIGURING, FAILED, TIMEOUT, PREEMPTED, and NODE_FAIL.
    #   %R: For pending jobs: the reason a job is waiting for execution is printed within parenthesis. For terminated jobs with failure: an  explanation  as to why the job failed is printed within parenthesis.  For all other job states: the list of allocate nodes.
    #   %r: reason a job is in its current state
    #   %u: username of the submitting user
    #   %U: numeric UID of the submitting user
    #
    def _stat_command(self, job):
        return "%s --noheader -o %%i^%%T^%%r -j %s" % \
            (self._squeue, job.lrms_jobid)

    def _parse_stat_output(self, stdout):
        """
        Receive the output of ``squeue --noheader -o %i:%T:%r and parse it.
        """
        jobstatus = dict()
        if stdout.strip() == '':
            # if stdout is empty and `squeue -j` exitcode is 0, then
            # the job has recently completed;
            #
            # if the job has been removed from the controllers'
            # memory, then `squeue -j` exits with code 1
            jobstatus['state'] = Run.State.TERMINATING
        else:
            # parse stdout
            jobid, state, reason = stdout.split('^')
            log.debug("translating SLURM's state '%s' to gc3libs.Run.State",
                      state)
            if state in ['PENDING', 'CONFIGURING']:
                # XXX: see above for a discussion of whether 'CONFIGURING'
                # should be grouped with 'RUNNING' or not; here it's
                # likely the correct choice to group it with 'PENDING' as
                # the "configuring" phase may last a few minutes during
                # which the job is not yet really running.
                jobstatus['state'] = Run.State.SUBMITTED
            elif state in ['RUNNING', 'COMPLETING']:
                jobstatus['state'] = Run.State.RUNNING
            elif state in ['SUSPENDED']:
                jobstatus['state'] = Run.State.STOPPED
            elif state in ['COMPLETED', 'CANCELLED', 'FAILED',
                           'NODE_FAIL', 'PREEMPTED', 'TIMEOUT']:
                jobstatus['state'] = Run.State.TERMINATING
            else:
                jobstatus['state'] = Run.State.UNKNOWN
        return jobstatus

    # acct cmd: sacct --noheader --parsable --format jobid,ncpus,cputimeraw,elapsed,submit,eligible,reserved,start,end,exitcode,maxrss,maxvmsize,totalcpu -j JOBID
    #
    # where:
    #   * jobid:      Job ID
    #   * alloccpus:  Count of allocated processors
    #   * cputimeraw: CPU time in seconds
    #   * elapsed:    Job duration, as [DD-][hh:]mm:ss
    #   * submit:     Time the job was submitted, as MM/DD-hh:mm:ss (timestamp in UTC) or ISO8601 format depending on compilation option
    #   * eligible:   When the job became eligible to run
    #   * reserved:   Difference between `start` and `eligible`
    #   * start:      Job start time, as MM/DD-hh:mm:ss (timestamp in UTC) or ISO8601 format depending on compilation option
    #   * end:        Termination time of the job, as MM/DD-hh:mm:ss (timestamp in UTC) or ISO8601 format depending on compilation option
    #   * exitcode:   exit code, ':', killing signal number
    #   * maxrss:     The maximum RSS across all tasks
    #   * maxvmsize:  The maximum virtual memory used across all tasks
    #   * totalcpu:   Total CPU time used by the job (does not include child processes, if any)
    #
    # Examples:
    #
    #    $ sudo sacct --noheader --parsable --format jobid,ncpus,cputimeraw,elapsed,submit,eligible,reserved,start,end,exitcode,maxrss,maxvmsize,totalcpu -j 14
    #    14|1|10|00:00:10|2012-09-23T23:38:41|2012-09-23T23:38:41|49710-06:28:06|2012-09-23T23:38:31|2012-09-23T23:38:41|0:0|||00:00:00|
    #
    #    $ sudo sacct --noheader --parsable --format jobid,ncpus,cputimeraw,elapsed,submit,start,end,exitcode,maxrss,maxvmsize,totalcpu -j 19
    #    19|1|66|00:01:06|2012-09-24T10:48:34|2012-09-24T10:47:28|2012-09-24T10:48:34|0:0|||00:00:00|
    #    19.0|1|65|00:01:05|2012-09-24T10:47:29|2012-09-24T10:47:29|2012-09-24T10:48:34|0:0|0|0|00:00:00|
    #
    #    $ env SLURM_TIME_FORMAT=standard sacct -S 0901 --noheader --parsable --format jobid,ncpus,cputimeraw,elapsed,submit,eligible,reserved,start,end,exitcode,maxrss,maxvmsize,totalcpu
    #    449002|32|128|00:00:04|2012-09-04T11:09:26|2012-09-04T11:09:38|2012-09-04T11:09:42|0:0|||00:00:00|
    #    449018|64|1472|00:00:23|2012-09-04T11:18:06|2012-09-04T11:18:24|2012-09-04T11:18:47|0:0|||00:01.452|
    #    449018.batch|1|23|00:00:23|2012-09-04T11:18:24|2012-09-04T11:18:24|2012-09-04T11:18:47|0:0|7884K|49184K|00:01.452|
    #    449057|3200|659200|00:03:26|2012-09-04T11:19:45|2012-09-04T11:19:57|2012-09-04T11:23:23|0:0|||00:01.620|
    #    449057.batch|1|206|00:03:26|2012-09-04T11:19:57|2012-09-04T11:19:57|2012-09-04T11:23:23|0:0|7896K|49184K|00:01.620|
    #
    # Warning: the `SLURM_TIME_FORMAT` environment variable influences how the times are reported,
    # so it should always be set to `standard` to get ISO8601 reporting.  See below for an example
    # of non-standard (SLURM_TIME_FORMAT=relative) report:
    #
    #    $ sacct --noheader --parsable --format jobid,ncpus,cputimeraw,elapsed,submit,eligible,reserved,start,end,exitcode,maxrss,maxvmsize,totalcpu -S 0901
    #    449002|32|128|00:00:04|4 Sep 11:09|4 Sep 11:09|00:00:12|4 Sep 11:09|4 Sep 11:09|0:0|||00:00:00|
    #    449018|64|1472|00:00:23|4 Sep 11:18|4 Sep 11:18|00:00:18|4 Sep 11:18|4 Sep 11:18|0:0|||00:01.452|
    #    449018.batch|1|23|00:00:23|4 Sep 11:18|4 Sep 11:18|00:00:-02|4 Sep 11:18|4 Sep 11:18|0:0|7884K|49184K|00:01.452|
    #    449057|3200|659200|00:03:26|4 Sep 11:19|4 Sep 11:19|00:00:12|4 Sep 11:19|4 Sep 11:23|0:0|||00:01.620|
    #    449057.batch|1|206|00:03:26|4 Sep 11:19|4 Sep 11:19|00:00:-02|4 Sep 11:19|4 Sep 11:23|0:0|7896K|49184K|00:01.620|
    #    449217|3200|0|00:00:00|4 Sep 11:42|4 Sep 11:42|00:00:05|4 Sep 11:43|4 Sep 11:43|0:0|||00:00.720|
    #
    # If SLURM accounting is disabled (default), then `sacct` outputs
    # an error message:
    #
    #    $ sacct 2
    #    SLURM accounting storage is disabled
    #
    def _acct_command(self, job):
        return ('env SLURM_TIME_FORMAT=standard %s --noheader --parsable' \
                ' --format jobid,exitcode,ncpus,elapsed,totalcpu,' \
                'submit,start,end,maxrss,maxvmsize -j %s' % \
                (self._sacct, job.lrms_jobid))

    def _parse_acct_output(self, stdout):
        acct = dict(
            exitcode=0,
            cores=0,
            duration=Duration(0, unit=seconds),
            used_cpu_time=Duration(0, unit=seconds),
            max_used_memory=Memory(0, unit=bytes))
        for line in stdout.split('\n'):
            line = line.strip()
            if line == '':
                continue
            # because of the trailing `|` we have an extra empty field
            jobid, exit, ncpus, elapsed, totalcpu, submit,\
                start, end, maxrss, maxvmsize, _ = line.split('|')
            # SLURM job IDs have the form `jobID[.step]`: only the
            # lines with the `step` part carry resource usage records,
            # whereas the total `jobID` line carries the exit codes
            # and overall duration/timing information.
            if '.' not in jobid:
                # master job record
                acct['duration'] = SlurmLrms._parse_duration(elapsed)
                acct['used_cpu_time'] = SlurmLrms._parse_duration(totalcpu)
                # compute POSIX exit status
                exitcode, signal = exit.split(':')
                acct['exitcode'] = (int(exitcode) << 8) + (int(signal) & 0x7f)
                # XXX: the master job record seems to report the
                # *requested* slots, whereas the step records report
                # the actual usage.  In our case these should be the
                # same, as the job script only runs one single step.
                # However, in the general case computing the *actual*
                # CPU usage is a mess, as we would have to check which
                # steps were executed simultaneously and which ones
                # were executed one after the other...
                acct['cores'] = int(ncpus)
                # provide starting point for resource usage records
                acct['max_used_memory'] = Memory(0, unit=MB)
                acct['slurm_max_used_ram'] = Memory(0, unit=MB)
                # XXX: apparently, Ubuntu's SLURM 2.3 has a bug
                # wherein `submit` == `end` in the master job record,
                # and the actual start time must be gathered from the
                # step records... try to work around
                submit = SlurmLrms._parse_timestamp(submit)
                start = SlurmLrms._parse_timestamp(start)
                end = SlurmLrms._parse_timestamp(end)
                acct['slurm_submission_time'] = min(submit, start)
                acct['slurm_start_time'] = end  # will be set when
                                                # looping on tasks,
                                                # see below
                acct['slurm_completion_time'] = max(submit, start, end)
            else:
                # common resource usage records (see Issue 78)
                vmem = SlurmLrms._parse_memspec(maxvmsize)
                acct['max_used_memory'] = max(vmem, acct['max_used_memory'])
                # SLURM-specific resource usage records
                mem = SlurmLrms._parse_memspec(maxrss)
                acct['slurm_max_used_ram'] = max(
                    mem, acct['slurm_max_used_ram'])
                # XXX: see above for timestamps
                submit = SlurmLrms._parse_timestamp(submit)
                start = SlurmLrms._parse_timestamp(start)
                acct['slurm_submission_time'] = min(
                    submit, acct['slurm_submission_time'])
                acct['slurm_start_time'] = min(start, acct['slurm_start_time'])
        return acct

    @staticmethod
    def _parse_duration(d):
        """
        Parse a SLURM duration expression, in the form ``DD-HH:MM:SS.UUU``.

        The ``DD``, ``HH`` and ``.UUU`` parts are optional.
        """
        total = Duration(0, unit=seconds)
        if '-' in d:
            # DD-HH:MM:SS
            ndays, d = d.split('-')
            total = Duration(int(ndays), unit=days)
        parts = list(reversed(d.split(':')))
        assert len(parts) > 0
        secs = parts[0]
        if '.' in secs:
            # SS.UUU
            total += Duration(float(secs), unit=seconds)
        else:
            total += Duration(int(secs), unit=seconds)
        if len(parts) > 1:
            total += Duration(int(parts[1]), unit=minutes)
        if len(parts) > 2:
            total += Duration(int(parts[2]), unit=hours)
        return total

    @staticmethod
    def _parse_memspec(m):
        unit = m[-1]
        if unit == 'G':
            return Memory(int(round(float(m[:-1]))), unit=GB)
        elif unit == 'M':
            return Memory(int(round(float(m[:-1]))), unit=MB)
        elif unit == 'K':  # XXX: not sure which one is used
            return Memory(int(round(float(m[:-1]))), unit=kB)
        else:
            # XXX: what does SLURM use as a default?
            return Memory(int(round(float(m))), unit=bytes)

    @staticmethod
    def _parse_timestamp(ts):
        """
        Parse a SLURM timestamp.

        The 'standard' format for SLURM timestamps is ISO8601;
        raise an error if any other format is detected.
        """
        # XXX: datetime.strptime() only available starting Py 2.5
        try:
            return datetime.datetime(
                *(time.strptime(ts, SlurmLrms._TIMEFMT_ISO8601)[0:6]))
        except ValueError, err:
            gc3libs.log.error(
                "Could not parse '%s' as an SLURM 'standard' (ISO8601)"
                " timestamp: %s: %s Please set the environment variable"
                " 'SLURM_TIME_FORMAT' to 'standard' on the SLURM frontend"
                " computer.", ts, err.__class__.__name__, err)
            # XXX: this results in an invalid timestamp...
            return None

    _TIMEFMT_ISO8601 = '%Y-%m-%dT%H:%M:%S'
    """
    A strptime() format string for parsing ISO8601 timestamps.
    """

    # kill cmd: scancel
    #
    # Just call it with the space-separated list of job IDs.
    #
    # - on successful cancellation, `scancel` emits no output::
    #
    #    $ scancel 2
    #
    # - for non-existing job IDs, `scancel` outputs an error message:
    #
    #    $ scancel 15
    #    scancel: error: Kill job error on job id 15: Invalid job id specified
    #
    def _cancel_command(self, jobid):
        return ("%s %s" % (self._scancel, jobid))

    @same_docstring_as(LRMS.get_resource_status)
    @LRMS.authenticated
    def get_resource_status(self):
        self.updated = False
        try:
            self.transport.connect()

            _command = ("%s --noheader -o '%%i^%%T^%%u^%%U^%%r^%%R'" %
                        self._squeue)
            log.debug("Running `%s`...", _command)
            exitcode, stdout, stderr = self.transport.execute_command(_command)
            if exitcode != 0:
                # cannot continue
                raise gc3libs.exceptions.LRMSError(
                    "SLURM backend failed executing '%s':"
                    " exit code: %d; stdout: '%s', stderr: '%s'"
                    % (_command, exitcode, stdout, stderr))

            log.debug("Computing updated values for total/available slots ...")
            (total_running, self.queued, self.user_run, self.user_queued) \
                = count_jobs(stdout, self._username)
            self.total_run = total_running
            self.free_slots = -1
            self.used_quota = -1

            log.info("Updated resource '%s' status:"
                     " free slots: %d,"
                     " total running: %d,"
                     " own running jobs: %d,"
                     " own queued jobs: %d,"
                     " total queued jobs: %d",
                     self.name,
                     self.free_slots,
                     self.total_run,
                     self.user_run,
                     self.user_queued,
                     self.queued,
                     )
            return self

        except Exception, ex:
            # self.transport.close()
            log.error("Error querying remote LRMS, see debug log for details.")
            log.debug("Error querying LRMS: %s: %s",
                      ex.__class__.__name__, str(ex), exc_info=True)
            raise


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="slurm",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
Test for classes and functions in the `utils` module.
"""
# Copyright (C) 2012, 2013, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'

# stdlib imports
import os
import sys
import unittest

# 3rd party imports
from nose.tools import assert_true, assert_false, assert_equal, raises
from nose.tools import set_trace
from nose.plugins.skip import SkipTest

# GC3Pie imports
import gc3libs.exceptions
import gc3libs.utils


## test definitions

class TestYieldAtNext(object):

    def test_YieldAtNext_yield(self):
        def generator_yield():
            yield 0
        g = gc3libs.utils.YieldAtNext(generator_yield())
        assert_equal(g.next(), 0)

    def test_YieldAtNext_send(self):
        def generator_yield_send():
            val = (yield 1)
            yield val
        g = gc3libs.utils.YieldAtNext(generator_yield_send())
        assert_equal(g.next(), 1)
        result = g.send('a sent value')
        assert_equal(result, None)
        assert_equal(g.next(), 'a sent value')

    def test_YieldAtNext_throw(self):
        def generator_yield_throw():
            try:
                val = (yield 2)
            except RuntimeError:
                yield 'exception caught'
        g = gc3libs.utils.YieldAtNext(generator_yield_throw())
        assert_equal(g.next(), 2)
        result = g.throw(RuntimeError)
        assert_equal(result, None)
        assert_equal(g.next(), 'exception caught')

    @raises(StopIteration)
    def test_YieldAtNext_StopIteration(self):
        def generator_yield():
            yield 3
        g = gc3libs.utils.YieldAtNext(generator_yield())
        assert_equal(g.next(), 3)
        # raises `StopIteration`
        g.next()


## main: run tests

if "__main__" == __name__:
    import nose
    nose.runmodule()



#! /usr/bin/env python
#
#   smd_projections.py -- Front-end script for submitting multiple
#   `smd_projections` jobs.
#
#   Copyright (C) 2013, 2014 GC3, University of Zurich
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `GSMD_projections` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``smd_projections --help`` for program usage instructions.
"""

__version__ = '2.2.1 version (SVN $Revision$)'
# summary of user-visible changes
__changelog__ = """

  2014-02-27:
  * Initial release, forked off the ``gstructure`` sources.
"""
__author__ = 'Tyanko Aleksiev <tyanko.aleksiev@chem.uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gsmd_projections
    gsmd_projections.GSMD_ProjectionsScript().run()

from pkg_resources import Requirement, resource_filename
import os
import posix
import tarfile

# gc3 library imports
import gc3libs.utils
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

## custom application class

class GSMD_ProjectionsApplication(Application):

    application_name = 'gsmd_projections'

    def __init__(self, **extra_args):
        """
        The wrapper script is being used for start the simulation. 
        """
        files_to_send = []

        smd_projections_wrapper_sh = resource_filename(Requirement.parse("gc3pie"),
                                              "gc3libs/etc/smd_projections_wrapper.sh")

        basename_input_tar = os.path.basename(extra_args['input_tar'])
        files_to_send.append((smd_projections_wrapper_sh,os.path.basename(smd_projections_wrapper_sh)))
        files_to_send.append((extra_args['input_tar'],basename_input_tar))
        

        cmd = "./smd_projections_wrapper.sh -d" 

        if 'calibration' in extra_args:
            cmd += " -b "
            files_to_send.append((extra_args['calibration'],'calibration.tar')) 

        cmd += " %s " % basename_input_tar

        cmd += " %s " % basename_input_tar.split('.')[0]

        extra_args['requested_memory'] = 6*GB

        self.output_dir = basename_input_tar.split('.')[0] + "_output" 
        extra_args['output_dir'] = self.output_dir

        Application.__init__(
            self,
            # arguments should mimic the command line interfaca of the command to be
            # executed on the remote end
            arguments = cmd,
            inputs = files_to_send,
            outputs = gc3libs.ANY_OUTPUT,
            stdout = 'smd_projections.log',
            join=True,
            **extra_args)

class GSMD_ProjectionsTask(RetryableTask, gc3libs.utils.Struct):
    """
    Run ``smd_projections`` on a given simulation directory until completion.
    """
    def __init__(self, **extra_args):
        RetryableTask.__init__(
            self,
            # actual computational job
            GSMD_ProjectionsApplication(**extra_args),
            # keyword arguments
            **extra_args)

## main script class

class GSMD_ProjectionsScript(SessionBasedScript):
    """
Scan the specified INPUT directories recursively for simulation
directories and submit a job for each one found; job progress is
monitored and, when a job is done, its output files are retrieved back
into the simulation directory itself.

The ``smd_projections`` command keeps a record of jobs (submitted, executed
and pending) in a session file (set name with the ``-s`` option); at
each invocation of the command, the status of all recorded jobs is
updated, output from finished jobs is collected, and a summary table
of all known jobs is printed.  New jobs are added to the session if
new input files are added to the command line.

Options can specify a maximum number of jobs that should be in
'SUBMITTED' or 'RUNNING' state; ``smd_projections`` will delay submission of
newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GSMD_ProjectionsTask,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GSMD_ProjectionsTask,
            )


    def setup_options(self):

        self.add_param("-b", "--calibration", metavar="CALIBRATION_DIR",
                       dest="calibration_dir", help="Use a different calibration directory.")

    def setup_args(self):

        self.add_param('input_source', type=str,
                       help="Projections input directory") 

    def parse_args(self):

        self.params.input_source = os.path.abspath(self.params.input_source)


    def new_tasks(self, extra):

        tasks = []
        input_tars = []
        extra_args = extra.copy()

        cwd = os.getcwd()
        os.chdir(self.params.input_source)    
             
        for projection_dir in self._list_local_folder(self.params.input_source):

            # Check if tar exists
            input_tar_file = projection_dir + ".tar"
            if os.path.isfile(input_tar_file):
                try:
                    os.remove(input_tar_file)
                except OSError, x:
                    gc3libs.log.error("Failed removing '%s': %s: %s",
                                      input_tar_file, x.__class__, x.message)
                    pass
            tar = tarfile.open(input_tar_file, "w:gz", dereference=True)
            tar.add(projection_dir)
            tar.close() 
            input_tars.append(tar)

        os.chdir(cwd)


        # Create an archive of the calibration directory
        if os.path.isdir(self.params.calibration_dir):

            calibration_tar_file = "calibration.tar"

            if os.path.isfile(calibration_tar_file):
                try:
                    os.remove(calibration_tar_file)
                except OSError, x:
                    gc3libs.log.error("Failed removing '%s': %s: %s",
                                      calibration_tar_file, x.__class__, x.message)
                    pass

            tar = tarfile.open(calibration_tar_file, "w:gz", dereference=True)
            tar.add(self.params.calibration_dir)
            tar.close()

            extra_args['calibration'] = calibration_tar_file

        for input_tar in input_tars:  

            jobname = "%s" % input_tar
            
            extra_args['jobname'] = jobname
    
            extra_args['input_tar'] = self.params.input_source + "/" + os.path.basename(input_tar.name) 
            
            self.log.info("Creating Task for input file: %s" % input_tar.name)
                
            tasks.append(GSMD_ProjectionsTask(
                **extra_args
                ))

        return tasks

    def _list_local_folder(self, input_folder):
        """
        Return a list of all the directories in the input folder.
        """
    
        return [ infile for infile in os.listdir(input_folder) if os.path.isdir(os.path.join(input_folder,infile)) ]





#! /usr/bin/env python
"""
Support for running a generic application with the GC3Libs.
"""
# Copyright (C) 2009-2012 GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '2.2.1 version (SVN $Revision: 3977 $)'


## main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="__init__",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011-2013, GC3, University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__docformat__ = 'reStructuredText'
__version__ = '$Revision$'

import datetime
import os
import shutil
import sys
import tempfile

import gc3libs
import gc3libs.core
import gc3libs.config
from gc3libs.backends.lsf import LsfLrms
from gc3libs.quantity import Duration, hours, minutes, seconds, Memory, GB, MB, kB

from nose.tools import assert_equal

_datetime_date = None

files_to_remove = []

def setUpModule():
    """Mock the `datetime.date.today()` outcome in order to make the LSF parsing independent from the testing date."""
    # save the original `datetime.date` to restore it in `tearDownModule()`
    import datetime
    global _datetime_date
    _datetime_date = datetime.date
    # mock features of `datetime.date.today()` that are actually used
    # in `LsfLrms._parse_date()`
    class MockDate(object):
        def __init__(self, real):
            self.__date = real
        def __getattr__(self, name):
            return getattr(self.__date, name)
        def __call__(self, *args, **kwargs):
            return self.__date(*args, **kwargs)
    datetime.date = MockDate(datetime.date)
    class Today(object):
        def __init__(self):
            self.year = 2012
            self.month = 12
    datetime.date.today = Today

def tearDownModule():
    # restore the original `datetime.date`
    global _datetime_date
    import datetime
    datetime.date = _datetime_date
    for fname in files_to_remove:
        if os.path.isdir(fname):
            shutil.rmtree(fname)
        else:
            os.remove(fname)

def test_get_command():
    (fd, tmpfile) = tempfile.mkstemp()
    files_to_remove.append(tmpfile)
    f = os.fdopen(fd, 'w+')
    f.write("""
[auth/ssh]
type=ssh
username=NONEXISTENT

[resource/example]
# mandatory stuff
type=lsf
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64

# alternate command paths
bsub = /usr/local/bin/bsub -R lustre
bjobs = /usr/local/bin/bjobs
lshosts = /usr/local/sbin/lshosts # comments are ignored!
""")
    f.close()

    cfg = gc3libs.config.Configuration()
    cfg.merge_file(tmpfile)
    b = cfg.make_resources()['example']

    assert_equal(b.bsub, ['/usr/local/bin/bsub', '-R', 'lustre'])

    assert_equal(b._bjobs,   '/usr/local/bin/bjobs')
    assert_equal(b._lshosts, '/usr/local/sbin/lshosts')

def test_bjobs_output_done1():
    jobstatus = LsfLrms._parse_stat_output("""
Job <131851>, Job Name <ChromaExtractShort>, User <wwolski>, Project <default>,
                     Status <DONE>, Queue <pub.8h>, Job Priority <50>, Command
                     <ChromatogramExtractor -in /cluster/scratch/malars/openswa
                     th/data/AQUA_fixed_water/split_napedro_L120224_001_SW-400A
                     QUA_no_background_2ul_dilution_10/split_napedro_L120224_00
                     1_SW-400AQUA_no_background_2ul_dilution_10_28.mzML.gz -tr
                     /cluster/scratch/malars/openswath/assays/iRT/DIA_iRT.TraML
                      -out split_napedro_L120224_001_SW-400AQUA_no_background_2
                     ul_dilution_10_28._rtnorm.chrom.mzML -is_swath -min_upper_
                     edge_dist 1 -threads 2>, Share group charged </lsf_biol_al
                     l/lsf_biol_other/wwolski>
Tue Jul 24 10:03:15: Submitted from host <brutus3>, CWD <$HOME/.gc3pie_jobs/lrm
                     s_job.YNZmU17755/.>, Output File <lsf.o%J>, Requested Reso
                     urces <select[mem<70000 && lustre] order[-ut] rusage[mem=1
                     000,m=1]>, Login Shell </bin/sh>, Specified Hosts <thin+9>
                     , <single+8>, <smp16+6>, <smp24+5>, <smp48+4>;

 RUNLIMIT
 480.0 min of a6122
Tue Jul 24 10:04:19: Started on <a6122>, Execution Home </cluster/home/biol/wwo
                     lski>, Execution CWD </cluster/home/biol/wwolski/.gc3pie_j
                     obs/lrms_job.YNZmU17755/.>;
Tue Jul 24 10:05:45: Done successfully. The CPU time used is 2.1 seconds.

 SCHEDULING PARAMETERS:
           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
 loadSched   -     -     -     -       -     -    -     -  1000M     -      -
 loadStop    -     -     -     -       -     -    -     -     -      -      -

          scratch      xs       s       m       l      xl      sp
 loadSched 4000.0      -       -       -       -       -       -
 loadStop      -       -       -       -       -       -       -
""")
    assert_equal(jobstatus.state, gc3libs.Run.State.TERMINATING)
    assert_equal(jobstatus.exit_status, 0)


def test_bjobs_output_done2():
    jobstatus = LsfLrms._parse_stat_output("""
Job <726659>, Job Name <Application>, User <wwolski>, Project <default>, Status
                      <DONE>, Queue <pub.8h>, Job Priority <50>, Command <FileM
                     erger -in /IMSB/users/wwolski/gc3pieScripts/split_napedro_
                     L120224_001_SW-400AQUA_no_background_2ul_dilution.ChromaEx
                     tractShort/split_napedro_L120224_001_SW-400AQUA_no_backgro
                     und_2ul_dilution_10_12._rtnorm.chrom.mzML /IMSB/users/wwol
                     ski/gc3pieScripts/split_napedro_L120224_001_SW-400AQUA_no_
                     background_2ul_dilution.ChromaExtractShort/split_napedro_L
                     120224_001_SW-400AQUA_no_background_2ul_dilution_10_13._rt
                     norm.chrom.mzML /IMSB/users/wwolski/gc3pieScripts/split_na
                     pedro_L120224_001_SW-400AQUA_no_background_2ul_dilution.Ch
                     romaExtractShort/split_napedro_L120224_001_SW-400AQUA_no_b
                     ackground_2ul_dilution_10_14._rtnorm.chrom.mzML /IMSB/user
                     s/wwolski/gc3pieScripts/split_napedro_L120224_001_SW-400AQ
                     UA_no_background_2ul_dilution.ChromaExtractShort/split_nap
                     edro_L120224_001_SW-400AQUA_no_background_2ul_dilution_10_
                     15._rtnorm.chrom.mzML /IMSB/users/wwolski/gc3pieScripts/sp
                     lit_napedro_L120224_001_SW-400AQUA_no_background_2ul_dilut
                     ion.ChromaExtractShort/split_napedro_L120224_001_SW-400AQU
                     A_no_background_2ul_dilution_10_16._rtnorm.chrom.mzML /IMS
                     B/users/wwolski/gc3pieScripts/split_napedro_L120224_001_SW
                     -400AQUA_no_background_2ul_dilution.ChromaExtractShort/spl
                     it_napedro_L120224_001_SW-400AQUA_no_background_2ul_diluti
                     on_10_17._rtnorm.chrom.mzML /IMSB/users/wwolski/gc3pieScri
                     pts/split_napedro_L120224_001_SW-400AQUA_no_background_2ul
                     _dilution.ChromaExtractShort/split_napedro_L120224_001_SW-
                     400AQUA_no_background_2ul_dilution_10_18._rtnorm.chrom.mzM
                     L /IMSB/users/wwolski/gc3pieScripts/split_napedro_L120224_
                     001_SW-400AQUA_no_background_2ul_dilution.ChromaExtractSho
                     rt/split_napedro_L120224_001_SW-400AQUA_no_background_2ul_
                     dilution_10_19._rtnorm.chrom.mzML /IMSB/users/wwolski/gc3p
                     ieScripts/split_napedro_L120224_001_SW-400AQUA_no_backgrou
                     nd_2ul_dilution.ChromaExtractShort/split_napedro_L120224_0
                     01_SW-400AQUA_no_background_2ul_dilution_10_1._rtnorm.chro
                     m.mzML /IMSB/users/wwolski/gc3pieScripts/split_napedro_L12
                     0224_001_SW-400AQUA_no_background_2ul_dilution.ChromaExtra
                     ctShort/split_napedro_L120224_001_SW-400AQUA_no_background
                     _2ul_dilution_10_20._rtnorm.chrom.mzML /IMSB/users/wwolski
                     /gc3pieScripts/split_napedro_L120224_001_SW-400AQUA_no_bac
                     kground_2ul_dilution.ChromaExtractShort/split_napedro_L120
                     224_001_SW-400AQUA_no_background_2ul_dilution_10_21._rtnor
                     m.chrom.mzML /IMSB/users/wwolski/gc3pieScripts/split_naped
                     ro_L120224_001_SW-400AQUA_no_background_2ul_dilution.Chrom
                     aExtractShort/split_napedro_L120224_001_SW-400AQUA_no_back
                     ground_2ul_dilution_10_22._rtnorm.chrom.mzML /IMSB/users/w
                     wolski/gc3pieScripts/split_napedro_L120224_001_SW-400AQUA_
                     no_background_2ul_dilution.ChromaExtractShort/split_napedr
                     o_L120224_001_SW-400AQUA_no_background_2ul_dilution_10_23.
                     _rtnorm.chrom.mzML /IMSB/users/wwolski/gc3pieScripts/split
                     _napedro_L120224_001_SW-400AQUA_no_background_2ul_dilution
                     .ChromaExtractShort/split_napedro_L120224_001_SW-400AQUA_n
                     o_background_2ul_dilution_10_24._rtnorm.chrom.mzML /IMSB/u
                     sers/wwolski/gc3pieScripts/split_napedro_L120224_001_SW-40
                     0AQUA_no_background_2ul_dilution.ChromaExtractShort/split_
                     napedro_L120224_001_SW-400AQUA_no_background_2ul_dilution_
                     10_25._rtnorm.chrom.mzML /IMSB/users/wwolski/gc3pieScripts
                     /split_napedro_L120224_001_SW-400AQUA_no_background_2ul_di
                     lution.ChromaExtractShort/split_napedro_L120224_001_SW-400
                     AQUA_no_background_2ul_dilution_10_26._rtnorm.chrom.mzML /
                     IMSB/users/wwolski/gc3pieScripts/split_napedro_L120224_001
                     _SW-400AQUA_no_background_2ul_dilution.ChromaExtractShort/
                     split_napedro_L120224_001_SW-400AQUA_no_background_2ul_dil
                     ution_10_27._rtnorm.chrom.mzML /IMSB/users/wwolski/gc3pieS
                     cripts/split_napedro_L120224_001_SW-400AQUA_no_background_
                     2ul_dilution.ChromaExtractShort/split_napedro_L120224_001_
                     SW-400AQUA_no_background_2ul_dilution_10_28._rtnorm.chrom.
                     mzML /IMSB/users/wwolski/gc3pieScripts/split_napedro_L1202
                     24_001_SW-400AQUA_no_background_2ul_dilution.ChromaExtract
                     Short/split_napedro_L120224_001_SW-400AQUA_no_background_2
                     ul_dilution_10_29._rtnorm.chrom.mzML /IMSB/users/wwolski/g
                     c3pieScripts/split_napedro_L120224_001_SW-400AQUA_no_backg
                     round_2ul_dilution.ChromaExtractShort/split_napedro_L12022
                     4_001_SW-400AQUA_no_background>, Share group charged </lsf
                     _biol_all/lsf_biol_other/wwolski>
Mon Jul 30 15:12:05: Submitted from host <brutus2>, CWD <$HOME/.gc3pie_jobs/lrm
                     s_job.QwyFmi4681/.>, Output File <lsf.o%J>, Requested Reso
                     urces <select[mem<70000] order[-ut] rusage[mem=1000,m=1]>,
                      Login Shell </bin/sh>, Specified Hosts <thin+9>, <single+
                     8>, <smp16+6>, <smp24+5>, <smp48+4>;

 RUNLIMIT
 480.0 min of a3168
Mon Jul 30 15:12:47: Started on <a3168>, Execution Home </cluster/home/biol/wwo
                     lski>, Execution CWD </cluster/home/biol/wwolski/.gc3pie_j
                     obs/lrms_job.QwyFmi4681/.>;
Mon Jul 30 15:12:56: Done successfully. The CPU time used is 1.7 seconds.

 SCHEDULING PARAMETERS:
           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
 loadSched   -     -     -     -       -     -    -     -     -      -      -
 loadStop    -     -     -     -       -     -    -     -     -      -      -

          scratch      xs       s       m       l      xl      sp
 loadSched20000.0      -       -       -       -       -       -
 loadStop      -       -       -       -       -       -       -
""")
    assert_equal(jobstatus.state, gc3libs.Run.State.TERMINATING)
    assert_equal(jobstatus.exit_status, 0)


def test_bjobs_output_exit_nonzero():
    jobstatus = LsfLrms._parse_stat_output("""
Job <132286>, User <wwolski>, Project <default>, Status <EXIT>, Queue <pub.1h>,
                     Job Priority <50>, Command <./x.sh>, Share group charged <
                     /lsf_biol_all/lsf_biol_other/wwolski>
Tue Jul 24 10:26:42: Submitted from host <brutus3>, CWD <$HOME/.>, Output File
                     <lsf.o%J>, Requested Resources <select[mem<70000] order[-u
                     t] rusage[mem=1024,xs=1]>, Specified Hosts <thin+9>, <sing
                     le+8>, <smp16+6>, <smp24+5>, <smp48+4>, <parallel+1>;
Tue Jul 24 10:26:47: Started on <a3010>, Execution Home </cluster/home/biol/wwo
                     lski>, Execution CWD </cluster/home/biol/wwolski/.>;
Tue Jul 24 10:26:53: Exited with exit code 42. The CPU time used is 0.0 seconds
                     .
Tue Jul 24 10:26:53: Completed <exit>.

 SCHEDULING PARAMETERS:
           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem
 loadSched   -     -     -     -       -     -    -     -     -      -      -
 loadStop    -     -     -     -       -     -    -     -     -      -      -

          scratch      xs       s       m       l      xl      sp
 loadSched20000.0      -       -       -       -       -       -
 loadStop      -       -       -       -       -       -       -
""")
    assert_equal(jobstatus.state, gc3libs.Run.State.TERMINATING)
    assert_equal(jobstatus.exit_status, 42)


def test_bacct_done0():
    """Test parsing accounting information of a <sleep 300> job."""
    # gotten with `bacct -l "jobid"`
    acct = LsfLrms._parse_acct_output("""
Accounting information about jobs that are:
  - submitted by all users.
  - accounted on all projects.
  - completed normally or exited
  - executed on all hosts.
  - submitted to all queues.
  - accounted on all service classes.
------------------------------------------------------------------------------

Job <3329613>, User <rmurri>, Project <default>, Status <DONE>, Queue <pub.1h>,
                     Command <sleep 60>, Share group charged </lsf_biol_all/lsf
                     _aeber/rmurri>
Mon Oct  8 17:07:54: Submitted from host <brutus4>, CWD <$HOME>, Output File <l
                     sf.o%J>;
Mon Oct  8 17:08:44: Dispatched to <a3201>;
Mon Oct  8 17:09:51: Completed <done>.

Accounting information about this job:
     Share group charged </lsf_biol_all/lsf_aeber/rmurri>
     CPU_T     WAIT     TURNAROUND   STATUS     HOG_FACTOR    MEM    SWAP
      0.08       50            117     done         0.0007     5M    222M
------------------------------------------------------------------------------

SUMMARY:      ( time unit: second )
 Total number of done jobs:       1      Total number of exited jobs:     0
 Total CPU time consumed:       0.1      Average CPU time consumed:     0.1
 Maximum CPU time of a job:     0.1      Minimum CPU time of a job:     0.1
 Total wait time in queues:    50.0
 Average wait time in queue:   50.0
 Maximum wait time in queue:   50.0      Minimum wait time in queue:   50.0
 Average turnaround time:       117 (seconds/job)
 Maximum turnaround time:       117      Minimum turnaround time:       117
 Average hog factor of a job:  0.00 ( cpu time / turnaround time )
 Maximum hog factor of a job:  0.00      Minimum hog factor of a job:  0.00

    """)
    assert_equal(acct['duration'], Duration('67s'))
    assert_equal(acct['used_cpu_time'], Duration('0.08s'))
    assert_equal(acct['max_used_memory'], Memory('227MB'))
    # timestamps
    year = datetime.date.today().year
    assert_equal(acct['lsf_submission_time'], datetime.datetime(year, 10, 8, 17, 7, 54))
    assert_equal(acct['lsf_start_time'],      datetime.datetime(year, 10, 8, 17, 8, 44))
    assert_equal(acct['lsf_completion_time'], datetime.datetime(year, 10, 8, 17, 9, 51))


def test_bacct_done1():
    """Test parsing `bacct -l` output for a not-so-trivial job."""
    acct = LsfLrms._parse_acct_output("""
Accounting information about jobs that are:
  - submitted by all users.
  - accounted on all projects.
  - completed normally or exited
  - executed on all hosts.
  - submitted to all queues.
  - accounted on all service classes.
------------------------------------------------------------------------------

Job <3329618>, User <rmurri>, Project <default>, Status <DONE>, Queue <pub.1h>,
                     Command <md5sum lsf.o3224113 lsf.o3224132>, Share group ch
                     arged </lsf_biol_all/lsf_aeber/rmurri>
Mon Oct  8 17:08:54: Submitted from host <brutus4>, CWD <$HOME>, Output File <l
                     sf.o%J>;
Mon Oct  8 17:10:01: Dispatched to <a3041>;
Mon Oct  8 17:10:07: Completed <done>.

Accounting information about this job:
     Share group charged </lsf_biol_all/lsf_aeber/rmurri>
     CPU_T     WAIT     TURNAROUND   STATUS     HOG_FACTOR    MEM    SWAP
      0.04       67             73     done         0.0005     3M     34M
------------------------------------------------------------------------------

SUMMARY:      ( time unit: second )
 Total number of done jobs:       1      Total number of exited jobs:     0
 Total CPU time consumed:       0.0      Average CPU time consumed:     0.0
 Maximum CPU time of a job:     0.0      Minimum CPU time of a job:     0.0
 Total wait time in queues:    67.0
 Average wait time in queue:   67.0
 Maximum wait time in queue:   67.0      Minimum wait time in queue:   67.0
 Average turnaround time:        73 (seconds/job)
 Maximum turnaround time:        73      Minimum turnaround time:        73
 Average hog factor of a job:  0.00 ( cpu time / turnaround time )
 Maximum hog factor of a job:  0.00      Minimum hog factor of a job:  0.00
    """)
    assert_equal(acct['duration'], Duration('6s'))
    assert_equal(acct['used_cpu_time'], Duration('0.04s'))
    assert_equal(acct['max_used_memory'], Memory('37MB'))
    # timestamps
    year = datetime.date.today().year
    assert_equal(acct['lsf_submission_time'], datetime.datetime(year, 10, 8, 17,  8, 54))
    assert_equal(acct['lsf_start_time'],      datetime.datetime(year, 10, 8, 17, 10,  1))
    assert_equal(acct['lsf_completion_time'], datetime.datetime(year, 10, 8, 17, 10,  7))


def test_bacct_killed():
    """Test parsing `bacct -l` output for a canceled job."""
    acct = LsfLrms._parse_acct_output("""
Accounting information about jobs that are:
  - submitted by all users.
  - accounted on all projects.
  - completed normally or exited
  - executed on all hosts.
  - submitted to all queues.
  - accounted on all service classes.
------------------------------------------------------------------------------

Job <3224113>, User <rmurri>, Project <default>, Status <EXIT>, Queue <pub.1h>,
                     Command <sleep 300>, Share group charged </lsf_biol_all/ls
                     f_aeber/rmurri>
Fri Oct  5 17:49:35: Submitted from host <brutus4>, CWD <$HOME>, Output File <l
                     sf.o%J>;
Fri Oct  5 17:50:35: Dispatched to <a3191>;
Fri Oct  5 17:51:30: Completed <exit>; TERM_OWNER: job killed by owner.

Accounting information about this job:
     Share group charged </lsf_biol_all/lsf_aeber/rmurri>
     CPU_T     WAIT     TURNAROUND   STATUS     HOG_FACTOR    MEM    SWAP
      0.04       60            115     exit         0.0003     1M     34M
------------------------------------------------------------------------------

SUMMARY:      ( time unit: second )
 Total number of done jobs:       0      Total number of exited jobs:     1
 Total CPU time consumed:       0.0      Average CPU time consumed:     0.0
 Maximum CPU time of a job:     0.0      Minimum CPU time of a job:     0.0
 Total wait time in queues:    60.0
 Average wait time in queue:   60.0
 Maximum wait time in queue:   60.0      Minimum wait time in queue:   60.0
 Average turnaround time:       115 (seconds/job)
 Maximum turnaround time:       115      Minimum turnaround time:       115
 Average hog factor of a job:  0.00 ( cpu time / turnaround time )
 Maximum hog factor of a job:  0.00      Minimum hog factor of a job:  0.00
""")
    assert_equal(acct['duration'], Duration('55s'))
    assert_equal(acct['used_cpu_time'], Duration('0.04s'))
    assert_equal(acct['max_used_memory'], Memory('35MB'))
    # timestamps
    year = datetime.date.today().year
    assert_equal(acct['lsf_submission_time'], datetime.datetime(year, 10, 5, 17, 49, 35))
    assert_equal(acct['lsf_start_time'],      datetime.datetime(year, 10, 5, 17, 50, 35))
    assert_equal(acct['lsf_completion_time'], datetime.datetime(year, 10, 5, 17, 51, 30))



# LSF incorporates resource usage information in a job's output;
# the job's output is a copy of the email that the LSF system
# sends to the user that submitted a job.
################################################################################################
# Sender: LSF System <lsfadmin@cpt086>
# Subject: Job 943186: <md5sum XXX/x.sh> Done
#
# Job <md5sum XXX/x.sh> was submitted from host <frt> by user <rmurri> in cluster <prdclst>.
# Job was executed on host(s) <cpt086>, in queue <normal>, as user <rmurri> in cluster <prdclst>.
# </home/rmurri> was used as the home directory.
# </home/rmurri> was used as the working directory.
# Started at Mon Oct  8 17:23:30 2012
# Results reported at Mon Oct  8 17:23:31 2012
#
# Your job looked like:
#
# ------------------------------------------------------------
# # LSBATCH: User input
# md5sum XXX/x.sh
# ------------------------------------------------------------
#
# Successfully completed.
#
# Resource usage summary:
#
#     CPU time   :      0.01 sec.
#     Max Memory :         2 MB
#     Max Swap   :        24 MB
#
#     Max Processes  :         1
#     Max Threads    :         1
#
# The output (if any) follows:
#
# b7f9d9c86469aa8b57c19af14bf80af9  XXX/x.sh
################################################################################################


if __name__ == "__main__":
    import nose
    nose.runmodule()



#! /usr/bin/env python
#
"""
Support for finding minima of functions with GC3Pie.

GC3Pie can run a large number of :class:`~gc3libs.Application` instances in
parallel. The idea of this optimization module is to use these core
capabilities to perform optimization, which is particularly effective for
optimization using evolutionary algorithms, as they require several independent
evaluations of the target function.

The optimization module has two main components, the driver and the algorithm.
You need both an instance of a driver and an instance of an algorithm to
perform optimization of a given function.

Drivers perform optimization following a specific algorithm. Two drivers are
currently implemented: :class:`drivers.SequentialDriver` that runs the entire
algorithm on the local computer (hence, all the evaluations of the target
function required by the algorithm are performed one after the other), and
:class:`drivers.ParallelDriver` splits the evaluations into tasks that are
executed in parallel using GC3Pie's remote execution facilities.

This module implements a generic framework for evolutionary algorithms, and one
particular type of global optimization algorithm called `Differential
Evolution`_ is worked out in full. Other Evolutionary Algorithms can easily be
incorporated by subclassing :class:`EvolutionaryAlgorithm`. (Different
optimization algorithms, for example gradient based methods such as
quasi-newton methods, could be implemented but likely require adaptations in
the driver classes.)

.. _`differential evolution`: http://stackoverflow.com/a/7519536

The module is organized as follows:

* :mod:`~gc3libs.optimizer.drivers`: Set of drivers that interface with GC3Libs 
  to automatically drive the optimization process following a specified
  algorithm. :class:`~gc3libs.optimizer.drivers.ParallelDriver` is the core of
  the optimization module, performing optimization using an algorithm based on
  :class:`EvolutionaryAlgorithm`.

* :mod:`~gc3libs.optimizer.dif_evolution`: Implements the Differential
  Evolution algorithm, in particular the evolution and selection step, based on
  :class:`EvolutionaryAlgorithm`. See the module for details on the algorithm.

* :mod:`~gc3libs.optimizer.extra`: Provides tools to printing, plotting etc. that can be
  used as addons to :class:`~gc3libs.optimizer.EvolutionaryAlgorithm`.

"""
# Copyright (C) 2011, 2012, 2013 University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__version__ = '$Revision$'
__author__ = 'Benjamin Jonen <benjamin.jonen@bf.uzh.ch>'
# summary of user-visible changes
__changelog__ = """

"""
__docformat__ = 'reStructuredText'

# For now use export export PYTHONPATH=~/workspace/globalOpt/gc3pie/ to allow import
# of gc3libs

# General imports
import os
import sys
import datetime
import glob
import time
import logging
import re
import shutil

import numpy as np

# gc3libs imports
import gc3libs
import gc3libs.debug
import gc3libs.config
import gc3libs.core
from gc3libs import Application, Run, Task, utils
from gc3libs.workflow import SequentialTaskCollection, ParallelTaskCollection

class EvolutionaryAlgorithm(object):
    '''
    Base class for building an evolutionary algorithm for global optimization.

    :param list initial_pop: Initial population for the optimization.
    :param int `itermax`: Maximum # of iterations.
    :param float `dx_conv_crit`: Abort optimization if all population members are within a certain distance to each other.
    :param float `y_conv_crit`: Declare convergence when the target function is below a `y_conv_crit`.
    :param obj `logger`: Configured logger to use.
    :param list `after_update_opt_state`: Functions that are called at the end of
                `update_opt_state`:meth:. Use this list
                to provide problem-specific printing and plotting routines. Examples can be found
                in `gc3libs.optimizer.extra`:mod:.
    '''

    def __init__(self, initial_pop,
                 # criteria for convergence
                 itermax = 100, dx_conv_crit = None, y_conv_crit = None,
                 # hooks for "extra" functions, e.g., printg/logging/plotting
                 logger=None, after_update_opt_state=[]):

        if logger:
            self.logger = logger
        else:
            self.logger = gc3libs.log

        # save parameters
        self.pop = np.array(initial_pop)
        self.dim = len(initial_pop[0])
        self.pop_size = len(initial_pop)

        self.y_conv_crit = y_conv_crit
        self.dx_conv_crit = dx_conv_crit

        self.itermax = itermax
        self.cur_iter = 0

        self.after_update_opt_state = after_update_opt_state


    def has_converged(self):
        '''
        Checks convergence based on two criteria:

        1) Is the lowest target value in the population below `y_conv_crit`.
        2) Are all population members within `dx_conv_crit` from the first population member.

        :rtype: bool
        '''
        converged = False
        # Check `y_conv_crit`
        if self.best_y < self.y_conv_crit:
            converged = True
            self.logger.info('Converged: self.best_y[%s] < self.y_conv_crit[%s]',
                             self.best_y, self.y_conv_crit)

        # Check `dx_conv_crit`
        dxs = np.abs(self.pop[:, :] - self.pop[0, :])
        has_dx_converged = (dxs <= self.dx_conv_crit).all()
        if has_dx_converged:
            converged = True
            self.logger.info('Converged: All population members within `dx_conv_crit` from the first population member. ')
        return converged


    def update_opt_state(self, new_pop, new_vals):
        '''
        Stores set of function values corresponding to the current
        population, then updates optimizer state in many ways:

        * update the `.best*` variables accordingly;
        * uses :meth:`select` to determine the surviving population.
        * advances iteration count.
        '''

        gc3libs.log.debug('Entering update_opt_state ...')
        # XXX: `new_vals` is a NumPy array, so the way it is printed
        # is influenced by Numpy's `set_printoptions()` -- the default
        # settings introduce line breaks at every 75th column, so the
        # following results in a multi-line log even for moderate-size
        # populations...  You might want to
        # `np.set_printoptions(linewidth=1024)` or so to prevent this.
        self.logger.debug('Updating optimizer state with new values: %s', str(new_vals))

        # In variable names `best` refers to a population member with the
        # lowest target function value within some group:

        # best_x: Coordinates of the best population member since the optimization started.
        # best_y: Val of the best population member since the optimization started.

        new_vals = np.array(new_vals)
        # determine the member with the lowest target value
        best_ix = np.argmin(new_vals)
        if self.cur_iter == 0 or new_vals[best_ix] < self.best_y:
            # store the best population members
            self.best_x = new_pop[best_ix, :].copy()
            self.best_y = new_vals[best_ix].copy()

        if self.cur_iter > 0:
            # update self.pop and self.vals
            self.select(new_pop, new_vals)
        else:
            self.pop = new_pop
            self.vals = new_vals

        self.logger.debug('Computed best value: %s (at index %d)', self.best_y, best_ix)

        for fn in self.after_update_opt_state:
            fn(self)

        self.cur_iter += 1

    def select(self, new_pop, new_vals):
        """
        Update `self.pop` and `self.vals` given the new population
        and the corresponding fitness vector.
        """
        raise NotImplemented(
            "Method `EvolutionaryAlgorithm.select` should be implemented in subclasses!")

    def evolve(self):
        '''
        Generates a new population fullfilling :func:`in_domain`.
        :rtype list of population members
        '''
        raise NotImplemented(
            "Method `EvolutionaryAlgorithm.evolve` should be implemented in subclasses!")


def populate(create_fn, in_domain=None, max_n_resample=100):
    '''
    Uses :func:`create_fn` to generate a new population. If :func:`in_domain` is not
    fulfilled, :func:`create_fn` is called repeatedly. Invalid population members are
    replaced until reaching the desired valid population size or
    `max_n_resample` calls to :func:`create_fn`. If `max_n_resample` is reached, a
    warning is issued and the optimization continues with the remaining
    "invalid" members.

    :param fun create_fn: Generates a new population. Takes no arguments.
    :param fun in_domain: Determines population's validity.
                          Takes no arguments and returns a list of bools
                          indicating each members validity.
    :param int max_n_resample: Maximum number of resamples to be drawn to
                               satisfy :func:`in_domain
                               
    :rtype: list of population members
    '''
    pop = create_fn()
    if in_domain:
        # re-evolve if some members do not fullfill fiter_fn
        pop_valid_orig = np.array(in_domain(pop))
        n_invalid_orig = (~pop_valid_orig).sum()
        fillin_pop = pop[~pop_valid_orig]
        n_to_fill = len(fillin_pop)
        total_filled = 0
        ctr = 0
        while total_filled < n_invalid_orig and ctr < max_n_resample:
            new_pop = create_fn()
            new_pop_valid = np.array(in_domain(new_pop))
            n_pop_valid = new_pop_valid.sum()
            new_total_filled = min(total_filled + n_pop_valid, n_to_fill)
            n_new_recruits = new_total_filled - total_filled
            ix_new_recruits = np.where(new_pop_valid)[0][0:n_new_recruits]
            fillin_pop[total_filled:new_total_filled] = new_pop[ix_new_recruits]
            total_filled = new_total_filled
        if total_filled < n_invalid_orig:
            self.logger.warning(
                "%d population members are invalid even after re-sampling %d times."
                "  You might want to increase `max_n_resample`.",
                (n_invalid_orig - total_filled), max_n_resample)
        pop[~pop_valid_orig] = fillin_pop
    return pop


def draw_population(lower_bds, upper_bds, dim, size, in_domain = None, seed = None):
    '''
    Draw a random population with the following criteria:

    :param list lower_bds: List of length `dim` indicating the lower bound in each dimension.
    :param list upper_bds: List of length `dim` indicating the upper bound in each dimension.
    :param int dim: Dimension of each population member.
    :param int size: Population size.
    :param fun in_domain: Determines population's validity.
                          Takes no arguments and returns a list of bools
                          indicating each members validity.
    :param float `seed`: Seed to initialize NumPy's random number generator.
    :rtype: list of population members
    '''
    np.random.seed(seed)
    return populate(create_fn=lambda:(lower_bds + np.random.random_sample( (size, dim) ) * ( upper_bds - lower_bds )),
                    in_domain=in_domain)




# -*- coding: utf-8 -*-

# pidlockfile.py
#
# Copyright © 2008–2009 Ben Finney <ben+python@benfinney.id.au>
#
# This is free software: you may copy, modify, and/or distribute this work
# under the terms of the Python Software Foundation License, version 2 or
# later as published by the Python Software Foundation.
# No warranty expressed or implied. See the file LICENSE.PSF-2 for details.

""" Lockfile behaviour implemented via Unix PID files.
    """

import os
import sys
import errno
import time

from gc3libs.compat.lockfile import (LockBase, AlreadyLocked, LockFailed, NotLocked,
                                     NotMyLock, LockTimeout)


class PIDLockFile(LockBase):
    """ Lockfile implemented as a Unix PID file.

    The lock file is a normal file named by the attribute `path`.
    A lock's PID file contains a single line of text, containing
    the process ID (PID) of the process that acquired the lock.

    >>> lock = PIDLockFile('somefile')
    >>> lock = PIDLockFile('somefile')
    """

    def __init__(self, path, threaded=False):
        # pid lockfiles don't support threaded operation, so always force
        # False as the threaded arg.
        LockBase.__init__(self, path, False)
        dirname = os.path.dirname(self.lock_file)
        basename = os.path.split(self.path)[-1]
        self.unique_name = self.path

    def read_pid(self):
        """ Get the PID from the lock file.
            """
        return read_pid_from_pidfile(self.path)

    def is_locked(self):
        """ Test if the lock is currently held.

            The lock is held if the PID file for this lock exists.

            """
        return os.path.exists(self.path)

    def i_am_locking(self):
        """ Test if the lock is held by the current process.

        Returns ``True`` if the current process ID matches the
        number stored in the PID file.
        """
        return self.is_locked() and os.getpid() == self.read_pid()

    def acquire(self, timeout=None):
        """ Acquire the lock.

        Creates the PID file for this lock, or raises an error if
        the lock could not be acquired.
        """

        end_time = time.time()
        if timeout is not None and timeout > 0:
            end_time += timeout

        while True:
            try:
                write_pid_to_pidfile(self.path)
            except OSError, exc:
                if exc.errno == errno.EEXIST:
                    # The lock creation failed.  Maybe sleep a bit.
                    if timeout is not None and time.time() > end_time:
                        if timeout > 0:
                            raise LockTimeout
                        else:
                            raise AlreadyLocked
                    time.sleep(timeout is not None and timeout/10 or 0.1)
                else:
                    raise LockFailed
            else:
                return

    def release(self):
        """ Release the lock.

            Removes the PID file to release the lock, or raises an
            error if the current process does not hold the lock.

            """
        if not self.is_locked():
            raise NotLocked
        if not self.i_am_locking():
            raise NotMyLock
        remove_existing_pidfile(self.path)

    def break_lock(self):
        """ Break an existing lock.

            Removes the PID file if it already exists, otherwise does
            nothing.

            """
        remove_existing_pidfile(self.path)

def read_pid_from_pidfile(pidfile_path):
    """ Read the PID recorded in the named PID file.

        Read and return the numeric PID recorded as text in the named
        PID file. If the PID file cannot be read, or if the content is
        not a valid PID, return ``None``.

        """
    pid = None
    try:
        pidfile = open(pidfile_path, 'r')
    except IOError:
        pass
    else:
        # According to the FHS 2.3 section on PID files in /var/run:
        # 
        #   The file must consist of the process identifier in
        #   ASCII-encoded decimal, followed by a newline character.
        # 
        #   Programs that read PID files should be somewhat flexible
        #   in what they accept; i.e., they should ignore extra
        #   whitespace, leading zeroes, absence of the trailing
        #   newline, or additional lines in the PID file.

        line = pidfile.readline().strip()
        try:
            pid = int(line)
        except ValueError:
            pass
        pidfile.close()

    return pid


def write_pid_to_pidfile(pidfile_path):
    """ Write the PID in the named PID file.

        Get the numeric process ID (“PID”) of the current process
        and write it to the named file as a line of text.

        """
    open_flags = (os.O_CREAT | os.O_EXCL | os.O_WRONLY)
    open_mode = 0644
    pidfile_fd = os.open(pidfile_path, open_flags, open_mode)
    pidfile = os.fdopen(pidfile_fd, 'w')

    # According to the FHS 2.3 section on PID files in /var/run:
    #
    #   The file must consist of the process identifier in
    #   ASCII-encoded decimal, followed by a newline character. For
    #   example, if crond was process number 25, /var/run/crond.pid
    #   would contain three characters: two, five, and newline.

    pid = os.getpid()
    line = "%(pid)d\n" % vars()
    pidfile.write(line)
    pidfile.close()


def remove_existing_pidfile(pidfile_path):
    """ Remove the named PID file if it exists.

        Removing a PID file that doesn't already exist puts us in the
        desired state, so we ignore the condition if the file does not
        exist.

        """
    try:
        os.remove(pidfile_path)
    except OSError, exc:
        if exc.errno == errno.ENOENT:
            pass
        else:
            raise



#! /usr/bin/env python
#
#   ggeosphere.py -- Front-end script for submitting multiple `GeoSphere` jobs.
#
#   Copyright (C) 2013, 2014 GC3, University of Zurich
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `GeoSphere` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``ggeosphere.py --help`` for program usage instructions.
"""

__version__ = '2.2.1 version (SVN $Revision$)'
# summary of user-visible changes
__changelog__ = """

  2014-03-21:
  * Initial release, forked off the ``ggeoshpere`` sources.
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@gc3.uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import ggeosphere
    ggeosphere.GeoSphereScript().run()

from pkg_resources import Requirement, resource_filename
import os
import posix
import logging
import ConfigParser
import time

# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

# Default values
TODAY=time.strftime('%Y%m%d-%H%M')
DEFAULT_S3CFG="$HOME/.s3cfg"
DEFAULT_CLOUD_ALL="*"
DEFAULT_MAX_RUNNING=10
DEFAULT_MAX_RETRIES=3
DEFAULT_LOG_LEVEL=logging.ERROR
# A4MESH_CONFIG_FILE="/etc/a4mesh/a4mesh.cfg"

## custom application class
def validate_config_option(config, option_name, default_value=None):
    if not config.has_option("default",option_name):
        if default_value:
            gc3libs.log.warning("No option '%s' defined. " \
                                "Using default: '%s'" % (option_name, str(default_value)))
            return default_value
        else:
            raise ConfigParser.ParsingError("No option '%s' defined." % option_name)
    else:
        return config.get('default',option_name)

class GeoSphereApplication(Application):
    """
    ``ggeosphere`` can work in two modes:
    1. CLI: as a regular SessionBasedScript can be invoked from the command like
    to process input model data. It runs as all SessionBasedScripts.
    2. Daemon: it runs continuously (like a daemon) and checks periodically for changes
    in the input/output folders. Every new model addedd to the input folder, triggers
    a new simualtion; every result removed from the output folder, triggers a re-run of the
    corresponding completed simualtion. Every model removed from the input folder also removes
    the corresponding result (unless overwritten by the ``PreserveOutputWhenInputRemoved``).
    """

    application_name = 'geosphere'

    def __init__(self, input_model, output_model, **extra_args):
        """
        Prepare remote execution of geosphere wrapper script.
        The resulting Application will associate a remote execution like:

        geosphere_wrapper.sh [options] <input model <output model>

        Options:
        -w <working dir>         name of the working dir extracted from 
                                 .tgz file
        -g <grok binary file>    path to 'grok' binary. Default in PATH
        -h <hgs binary file>     path to 'hgs' binary. Default in PATH
        -d                       enable debug
        """


        self.input_model = input_model
        self.output_model = output_model

        inputs = []
        outputs = []

        geosphere_wrapper_sh = resource_filename(
            Requirement.parse("gc3pie"),"gc3libs/etc/geosphere_wrapper.sh"
        )

        inputs.append((geosphere_wrapper_sh,
                       os.path.basename(geosphere_wrapper_sh)))

        cmd = "./geosphere_wrapper.sh -d "

        if 's3cfg' in extra_args:
            inputs.append((extra_args['s3cfg'],
                           "etc/s3cfg"))

        if 'grok_bin' in extra_args:
            cmd += "-g %s " % extra_args['grok_bin']

            inputs.append((extra_args['grok_bin'],
                          os.path.join("./bin",
                                       os.path.basename(extra_args['grok_bin']))
                       ))

        if 'hgs_bin' in extra_args:
            cmd += "-h %s " % extra_args['hgs_bin']

            inputs.append((extra_args['hgs_bin'],
                          os.path.join("./bin",
                                       os.path.basename(extra_args['hgs_bin']))
                       ))

        # Pass input model location
        if self.input_model.scheme == 'file':
            # Include local input files as part of Application
            inputs.append((str(self.input_model),os.path.join(".",
                                                              os.path.basename(
                                                                  str(self.input_model)))))
            cmd += " ./%s " % os.path.basename(str(self.input_model))
        else:
            # just pass remote input model URI
            cmd += "%s" % str(input_model)

        # Pass working dir name argument
        cmd += " %s " % extra_args['jobname']

        # Pass output location argument
        if self.output_model.scheme == 'file':
            # Include output as part of data to be retrieved locally
            outputs.append((os.path.basename(str(self.output_model)),
                            str(self.output_model)))
            cmd += "./%s " % os.path.basename(str(self.output_model))
        else:
            # just pass remote output model URI
            cmd += "%s " % str(self.output_model)

        Application.__init__(
             self,
             # arguments should mimic the command line interface of the command to be
             # executed on the remote end
             arguments = cmd,
             inputs = inputs,
             outputs = outputs,
             stdout = 'geosphere.log',
             join=True,
             **extra_args)


        # #XXX: TESTING PURPOSES...
        # extra_args['requested_memory'] = 1*MB

        # Application.__init__(
        #     self,
        #     # the following arguments are mandatory:
        #     arguments = ["/bin/hostname"],
        #     inputs = [],
        #     outputs = [],
        #     stdout = "stdout.txt",
        #     join = True,
        #     **extra_args)

class GeoSphereTask(RetryableTask, gc3libs.utils.Struct):
    """
    Run ``geosphere`` on a given simulation directory until completion.
    """
    def __init__(self, input_model, output_model, **extra_args):
        RetryableTask.__init__(
            self,
            # actual computational job
            GeoSphereApplication(input_model, output_model, **extra_args),
            # keyword arguments
            **extra_args)

## main script class

class GeoSphereScript(SessionBasedScript):
    """
    Loads the configuration file.
    Scans the specified INPUT for model files to simulate with
    ggeosphere. Submit a job for each one found; job progress is
    monitored and, when a job is done, its output files are retrieved back
    into the simulation directory itself.
    
    The ``ggeosphere`` has two working models: classical CLI based that can be executed
    as local script or as web-based application (use the '--daemon-mode' to enable
    web-based application mode). In web-based application mode ``ggeosphere``
    does not terminate when all input models have been processed but it 
    periodically checks whether new models have been added.
    In case, a new simulation is launched.
    It keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.  

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``ggeosphere`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GeoSphereTask,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GeoSphereTask,
            )

    def setup_options(self):
        self.add_param("-I", "--Input", 
                       metavar="PATH",
                       action="store",
                       default=None,
                       dest="input_folder", 
                       help="Path to the input folder. "+
                       "Valid URLs: [local path], [s3://].")

        self.add_param("-O", "--Output",
                       metavar="PATH",
                       action="store",
                       default=None,
                       dest="output_folder", 
                       help="Path to the output folder. "+
                       "Valid URLs: [local path], [s3://].")

        self.add_param("-c", "--config", 
                       metavar="PATH",
                       action="store",
                       default=None,
                       dest="config_file", 
                       help="Location of the configuration file. "+
                       "Command line options take precedence over config file.")

        self.add_param("-G", "--grok", metavar="PATH",
                       dest="grok_bin", default=None,
                       help="Location of the 'grok' binary. "+
                       "Default: assumes 'grok' binary being "+
                       "available on the remote execution node.")

        self.add_param("-H", "--hgs", metavar="PATH",
                       dest="hgs_bin", default=None,
                       help="Location of the 'hgs' binary. "+
                       "Default: assumes 'hgs' binary being "+
                       "available on the remote execution node.")

        self.add_param("-Y", "--s3cfg", metavar="PATH",
                       dest="s3cfg", default=None,
                       help="Location of the s3cfg configuration file. "+
                       "Default: assumes 's3cfg' configuration file being "+
                       "available on the remote execution node.")

        self.add_param("-P", "--preserve-output", metavar="BOOLEAN",
                       dest="preserve_output", default=True,
                       help="When an input model is removed, the corresponding "+
                       "output files will be kept. Default: False (remove output).")

        self.add_param("-d", "--daemon-mode", metavar="BOOLEAN",
                       dest="is_daemon", default=False,
                       help="Run ggeosphere in daemon mode. Default: False (no daemon).")

    def parse_args(self):
        """
        Check presence of input folder (should contains R scripts).
        path to command_file should also be valid.
        """

        # Check configuration file
        if self.params.config_file:
            if not os.path.isfile(self.params.config_file):
                raise gc3libs.exceptions.InvalidUsage("Config file '%s' not found" % self.params.config_file)

            self._read_config_file(self.params.config_file)

        if self.params.input_folder:
            # Overwrite configuration file
            self.extra['input'] = self.params.input_folder

        if self.params.output_folder:
            # Overwrite configuration file
            self.extra['output'] = self.params.output_folder
        
        if not self.extra['input']:
            raise gc3libs.exceptions.InvalidUsage(
                "Missing required argument 'input_folder'. " \
                "Either set in configuration file as " \
                "'InputLocation' or using the '--input' option.")

        if not self.extra['output']:
            raise gc3libs.exceptions.InvalidUsage(
                "Missing required argument 'output_folder'. " \
                "Either set in configuration file as "
                "'OutputLocation' or using the '--input' option.")

        # Transform input and output references in Url
        if not self.extra['input'].endswith('/'):
            self.extra['input'] += '/'
        self.extra['input'] = gc3libs.url.Url(self.extra['input'])

        if not self.extra['output'].endswith('/'):
            self.extra['output'] += '/'
        self.extra['output'] = gc3libs.url.Url(self.extra['output'])

        # Read command line options
        if self.params.grok_bin:
            if not os.path.isfile(self.params.grok_bin):
                raise gc3libs.exceptions.InvalidUsage(
                    "grok binary '%s' does not exists"
                    % self.params.grok_bin)
            self.extra['grok'] = self.params.grok_bin
            self.log.info("Grok binary: [%s]" % self.extra['grok'])
        else:
            self.log.info("Grok binary: [use remote]")

        if self.params.hgs_bin:
            if not os.path.isfile(self.params.hgs_bin):
                raise gc3libs.exceptions.InvalidUsage(
                    "hgs binary '%s' does not exists"
                    % self.params.hgs_bin)
            self.extra['hgs'] = self.params.hgs_bin
            self.log.info("hgs binary: [%s]" % self.extra['hgs'])
        else:
            self.log.info("hgs binary: [use remote]")

        if self.params.s3cfg:
            if not os.path.isfile(self.params.s3cfg):
                raise gc3libs.exceptions.InvalidUsage(
                    "s3cfg config file '%s' does not exists"
                    % self.params.s3cfg)
            self.extra['s3cfg'] = self.params.s3cfg
            self.log.info("s3cfg config: [%s]" % self.extra['s3cfg'])
        else:
            self.log.info("s3cfg config: [use remote]")

        self.extra['is_daemon'] = self.params.is_daemon

        # Overwrite default only if set to False
        if self.params.preserve_output == False:
            self.extra['preserve_output'] = False

        # Get input/output folder URL schema
        if self.extra['input'].scheme.lower() == "s3" or self.extra['output'].scheme.lower() == "s3":
            # Then check validity of s3cmd configuration file
            if not os.path.isfile(self.extra['s3cfg']):
                raise gc3libs.exceptions.InvalidArgument("S3cmd config file '%s' not found" \
                                                         % self.extra['s3cfg'])
        if self.extra['input'].scheme.lower() == "file":
            # Check local filesystem
            if not os.path.exists(os.path.abspath(self.extra['input'].path)):
                raise gc3libs.exceptions.InvalidArgument("Local input folder %s not found" \
                                                         % os.path.abspath(self.extra['input'].path))
        if self.extra['output'].scheme.lower() == "file":
            # Check local folder; create it in case
            if not os.path.exists(os.path.abspath(self.extra['output'].path)):
                gc3libs.log.info("Creating output folder '%s'", self.extra['output'].path)
                os.makedirs(self.extra['output'].path)
            

        self.log.info("Input models location: [%s]" % str(self.extra['input']))
        self.log.info("Output results location: [%s]" % str(self.extra['output']))
        self.log.info("Running as daemon: [%s]" % str(self.extra['is_daemon']))
        


    def _create_task(self, input_model, output_model, **extra_args):
        
        # XXX: Weak
        # Convention is that input model files are named after the folder
        # they represent. So an input model:
        # "emme_v2_20131125-lowres-all2012_365d_monthly_DKpaper_LINUX.tgz"
        # will produce, once expanded, a folder named:
        # emme_v2_20131125-lowres-all2012_365d_monthly_DKpaper_LINUX
        working_dir = os.path.splitext(os.path.basename(str(input_model)))[0]
        
        extra_args['jobname'] = working_dir
        
        # FIXME: ignore SessionBasedScript feature of customizing 
        # output folder
        # extra_args['output_dir'] = self.params.output
        extra_args['output_dir'] = os.path.join(self.extra['simulation_log_location'],extra_args['jobname'])

        extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', extra_args['jobname'])
        extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', extra_args['jobname'])

        extra_args['s3cfg'] = self.extra['s3cfg']
            
        self.log.info("Creating Task for input file: %s" % os.path.basename(str(input_model)))
        
        return GeoSphereTask(
            input_model,
            output_model,
            **extra_args
        )
        

    def new_tasks(self, extra):

        self.tasks = []

        for input_model in self._list_models_by_url(self.extra['input']):
            self.tasks.append(self._create_task(input_model, self._get_output(input_model), **extra.copy()))

        return self.tasks

    def every_main_loop(self):
        # Check for new tasks
        # eventually add them to self.controller

        gc3libs.log.debug("Reading S3 folder for new models")

        input_list = self._list_models_by_url(self.extra['input'])
        output_list = self._list_models_by_url(self.extra['output'])

        # for task in self.session:
        #     if task.execution.state in [Run.State.TERMINATED]:
        #         # Check whether output is present
        #         if task.output_model
                


        for input_model in self._list_models_by_url(self.extra['input']):
            # check if input_model is already associated to a running Task

            # Check for input not yet processed: create new task
            # Check for running tasks with missing input: terminate task
            # Check for missing output: re-start task (from TERMIANTED to NEW)

            skip = False

            # skip those 'input_model' that either have been already
            # processed or are already associated to a running task
            for task in self.session:
                if task.input_model == input_model:
                    skip = True
                    break

            if not skip:
                # Add as a new task
                gc3libs.log.info("Found new model to add: '%s'" % os.path.basename(str(input_model)))
                task =  self._create_task(input_model, self._get_output(input_model), **self.extra.copy())
                self.session.add(task)
                self._controller.add(task)

        # # Remove all terminated tasks
        # for task in self.session.tasks:
        #     if task.state in [Run.State.TERMINATED]:
        #         # Remove task from controller
        #         # This should help free memory
        #         self.controlled.remove(task)
        #         self.session.remove(task.id)


    def _get_output(self, input_model):
        """
        Returns an output gc3libs.url.Url representing
        the expected output file procudes at the end of the 
        GeoSphere simulation given the provided input_model.
        Agreed convention with the A4Mesh team:
        output: 'result_<input_model>_<date>.tgz'
        """
        return gc3libs.url.Url(os.path.join(str(self.extra['output']),
                                            "result_%s_%s.tgz" % 
                                            (os.path.splitext(
                                                os.path.basename(
                                                    str(input_model)))[0],
                                             TODAY)))

    def _list_models_by_url(self, url):
        """
        """

        if url.scheme == "s3":
            return self._list_s3(url)
        elif url.scheme ==  "file":
            return self._list_local(url)
        else:
            gc3libs.log.error("Unsupported Input folder URL %s. "+
                              "Only supported protocols are: [file,s3] " % url.scheme)
            return None

    def _list_local(self, input_folder):
        """
        return a list of all .fastq files in the input folder
        """
    
        return [ gc3libs.url.Url(os.path.join(input_folder.path,infile)) for infile in os.listdir(input_folder.path) if infile.endswith('.tgz') ]

    def _list_s3(self, s3_url):
        """
        Use s3cmd command line interface to interact with
        a remote S3-compatible ObjectStore.
        Assumption: 
        . s3cmd configuration file available
        and correctly pointing to the right ObjectStore.
        . s3cmd available in PATH environmental variable.
        . Valid for only 1 S3_URL path
        """

        import subprocess

        # read content of remote S3CMD_URL
        # 's3cmd ls' should return a list of model archives
        # for each of them bundle archive_name and working_dir
        _command = "s3cmd "
        if self.extra['s3cfg']:
            _command += " -c %s " % self.extra['s3cfg']
            
        _command += " ls %s"  % str(s3_url)

        try:
            _process = subprocess.Popen(_command,
                                        stdout=subprocess.PIPE, 
                                        stderr=subprocess.PIPE,
                                        close_fds=True, shell=True)
                
            (out, err) = _process.communicate()
            exitcode = _process.returncode
            
            if (exitcode != 0):
                gc3libs.log.error("Failed accessing S3 folder. "
                                  "Exitcode %s, error %s" % (exitcode,err))
                yield None
        except Exception, ex:
            gc3libs.log.error("Failed while reading remote S3 folder: "+
                              "%s", str(ex))
            yield None


        # Parse 'out_list' result and extract all .tgz or .zip archive names

        for s3_obj in out.strip().split("\n"):
            if s3_obj.startswith("DIR"):
                # it's a S3 directory; ignore
                continue
            # object string format: '2014-01-09 16:41 3627374 s3://a4mesh/model_1.tgz'
            try:
                s3_url = s3_obj.split()[3] 
                if(s3_url.startswith("s3://")) and(s3_url.endswith("tgz")):
                    yield gc3libs.url.Url(s3_url)
            except IndexError:
                gc3libs.log.warning("Unexpected model file: '%s'. " \
                                    "Ingoring error." % s3_obj)
                continue


    def _main_loop_exitcode(self, stats):
        """
        This works in daemon mode;
        Thus do not end if all simulations have been
        processed. Continue and wait for new models to be uploaded.
        """
        if self.extra['is_daemon']:
            return 4
        else:
            return SessionBasedScript._main_loop_exitcode(self, stats)

    def _read_config_file(self, config_file):

        config = ConfigParser.RawConfigParser()
        config.read(self.params.config_file)
        if not config.has_section("default"):
            raise ConfigParser.ParsingError("Section 'default' not found")

        self.extra['s3cfg'] = os.path.expandvars(validate_config_option(config, 
                                                                        "S3ConfigFileLocation", 
                                                                        DEFAULT_S3CFG))

        self.extra['input'] = os.path.expandvars(validate_config_option(config, 
                                                                        "InputLocation"))
        
        self.extra['output'] = os.path.expandvars(validate_config_option(config, 
                                                                         "OutputLocation"))

        self.extra['preserve-output'] = validate_config_option(config, 
                                                               "PreserveOutputWhenInputRemoved")


        self.extra['computing_resources'] = validate_config_option(config, 
                                                                   "computing_resources", 
                                                                   DEFAULT_CLOUD_ALL)

        self.extra['max_running'] = validate_config_option(config, 
                                                           "max_running", 
                                                           DEFAULT_MAX_RUNNING)

        self.extra['max_retries'] = validate_config_option(config, 
                                                           "max_retries", 
                                                           DEFAULT_MAX_RETRIES)

        self.extra['log_level'] = validate_config_option(config, 
                                                         "log_level", 
                                                         DEFAULT_LOG_LEVEL)

        self.extra['save_simulation_log'] = validate_config_option(config, 
                                                                   "save_simulation_log", 
                                                                   False)

        self.extra['simulation_log_location'] = os.path.expandvars(validate_config_option(config, 
                                                                       "simulation_log_location", 
                                                                       False))


###  TODO  ####
# 
# . Wrapper script review
# . pass also 'working_dir' name to wrapper script
# . error in output_file location (distinguish between local vs S3)
# . error in output file location for wrapper

# wrapper script should be invoked as follow:
# ./geosphere_wrapper.sh -d s3://models/input/emme_v2_20131125-lowres-all2011_365d_monthly_DKpaper_LINUX.tgz emme_v2_20131125-lowres-all2011_365d_monthly_DKpaper_LINUX s3://models/output/result_emme_v2_20131125-lowres-all2011_365d_monthly_DKpaper_LINUX_20140507-1612.tgz



